/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.image.ImageObserver;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Vector;

import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTable.PrintMode;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.RepaintManager;
import javax.swing.event.TableModelListener;
import javax.xml.stream.XMLStreamException;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodicMatrix;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;

/**
 * @author Christoph Halbmayr
 */
class EditorListPanel extends EditorChangePanel implements Printable, EditorPanelInterface {

	/** Contains Toolbar buttons */
	private Vector<JButton> Buttons;

	/** Contains the titles */
	private boolean newVarRow = false;

	/** selected row */
	private int selectedRow;

	/** name of the Selected TStruct */
	private String selectedStructName;

	/** Contains the struct */
	private TStruct struct;

	/** MainTable Action Lsitener */
	ActionListener al;

	/** Contains Table Titels and Table Data */
	TableModelListPanel model;

	/** JPanel */
	JPanel panel;

	/** Signal Editor Table */
	JTable table;

	/** ToolBar for Navigation und editing Signals */
	JToolBar tb;

	/** Contains the titles */
	Vector<String> titles;

	/**
	 * Creats an empty Object
	 */
	public EditorListPanel() {

		this.initialize();
	}

	/**
	 * Adds Action, Key and Mouse Listener to the Table
	 * 
	 * @param a
	 *            ActionListener
	 * @param kl
	 *            KeyListener
	 * @param ml
	 *            MouseListener
	 */
	public void addListener(final ActionListener a, final KeyListener kl, final MouseListener ml) {
		this.table.addKeyListener(kl);
		this.table.addMouseListener(ml);
		for (final JButton i : this.Buttons) {
			i.addActionListener(a);
		}

	}

	/**
	 * creats a new TData
	 */
	public void addNewVar() {
		final int row = this.table.getSelectedRow();
		final String name = this.model.getValueAt(row, 0).toString();
		final String datatype = this.model.getValueAt(row, 1).toString();

		if ((name.equals("[NAME]") == false) && (datatype.equals("[Datatype]") == false)) {

			this.struct.put(name, this.newTData(datatype));
			this.reload();

		}

	}

	/**
	 * Add a data to the table
	 * 
	 * @param name
	 *            Name of data
	 * @param data
	 *            Data to add
	 */
	public void addData(String name, TData data) {
		this.removeNewVarRow();

		this.model.addData(name, data);

		this.struct.put(name, data);
		this.model.getDataVector().removeAllElements();
		this.addTStruct(this.struct);

		this.columsResize(this.table, 90);
		this.reload();
		this.addNewVarRow();
	}

	@Override
	public void addTableModelListener(final TableModelListener tml) {

		this.table.getModel().addTableModelListener(tml);
	}

	/**
	 * Adds a TStruct to the Editor
	 * 
	 * @param struct
	 *            TStruct
	 */
	public void addTStruct(final TStruct struct) {

		for (final String s : struct.getKeys()) {
			this.model.addData(s, struct.get(s));
		}
		this.struct = struct;
		this.columsResize(this.table, 90);
		this.addNewVarRow();
	}

	@Override
	public void changed(final int row, final int column) {
		if (row == (this.table.getRowCount() - 1)) {
			this.addNewVar();
		} else if (column == 0) {

			final String newName = this.model.getValueAt(row, column).toString();
			final String oldName = this.struct.getKeys()[row];
			final TData data = this.struct.get(oldName);

			this.struct.remove(oldName);
			this.struct.put(newName, data);
			this.reload();

		} else if (column == 1) {
			final String key = this.model.getValueAt(row, 0).toString();
			final String datatype = this.model.getValueAt(row, 1).toString();
			final String oldDatatype = this.struct.get(key).getClass().getName().split("\\.")[5];

			if (datatype.endsWith(oldDatatype) == false) {
				this.struct.remove(key);
				this.struct.put(key, this.newTData(datatype));
				this.reload();
			}
		}
	}

	/**
	 * Deletes the selectet row and the TData
	 */
	@Override
	public void delete() {

		final int[] rowsToDelete = this.table.getSelectedRows();

		for (final int i : rowsToDelete) {
			String key = this.model.getValueAt(i, 0).toString();
			key = key.replaceAll(" ", "");
			this.struct.remove(key);
		}

		this.reload();

	}

	@Override
	public JPanel getPanel() {
		return this.panel;

	}

	/**
	 * @return selected Struct Name
	 */
	public String getSelectedStructName() {

		return this.selectedStructName;

	}

	/**
	 * Adds a Signal to the table
	 * 
	 * @return TData
	 */
	@Override
	public TData getSelectedTData() {

		if (this.model.getValueAt(this.table.getSelectedRow(), 0) != null) {
			String key = this.model.getValueAt(this.table.getSelectedRow(), 0).toString();
			key = key.replaceAll(" ", "");
			this.selectedStructName = this.model.getValueAt(this.table.getSelectedRow(), 0).toString();
			this.selectedRow = this.table.getSelectedRow();
			return this.struct.get(key);

		} else {
			return null;
		}
	}

	@Override
	public TData getTData() {
		return this.struct;
	}

	@Override
	public void importTData(final File f) {
		TStruct temp = null;

		try {
			temp = (TStruct) TData.read(new FileInputStream(f));
		} catch (final FileNotFoundException e) {
			e.printStackTrace();
		} catch (final TDataException e) {
			e.printStackTrace();
		} catch (final XMLStreamException e) {
			e.printStackTrace();
		} catch (final IOException e) {
			e.printStackTrace();
		}

		this.model.getDataVector().removeAllElements();

		this.struct = temp;
		this.addTStruct(this.struct);

		this.reload();
	}

	@Override
	public int print(final Graphics g, final PageFormat pageFormat, final int pageIndex) throws PrinterException {

		if (pageIndex > 0) {
			return Printable.NO_SUCH_PAGE;
		} else {
			final int x = (int) pageFormat.getImageableX() + 1;
			final int y = (int) pageFormat.getImageableY() + 1;
			g.translate(x, y);
			final RepaintManager currentManager = RepaintManager.currentManager(this.panel);

			currentManager.setDoubleBufferingEnabled(true);

			final JTable printTable = this.table;
			final TableModelListPanel m = (TableModelListPanel) printTable.getModel();
			m.removeRow(m.getRowCount() - 1);

			printTable.paintComponents(g);
			currentManager.setDoubleBufferingEnabled(true);
			return Printable.PAGE_EXISTS;
		}

	}

	@Override
	public void printme() {
		final PrinterJob printJob = PrinterJob.getPrinterJob();
		final MessageFormat headerFormat = new MessageFormat("Page {0}");
		final MessageFormat footerFormat = new MessageFormat("- {0} -");

		final JTable printTable = this.table;
		final TableModelListPanel m = (TableModelListPanel) printTable.getModel();
		m.removeRow(m.getRowCount() - 1);

		printJob.setPrintable(printTable.getPrintable(PrintMode.FIT_WIDTH, headerFormat, footerFormat));
		if (printJob.printDialog()) {
			try {
				printJob.print();

			} catch (final PrinterException pe) {
			}
		}

	}

	/**
	 * Reloads the whole contetnt of the Table
	 */
	@Override
	public void reload() {

		this.newVarRow = false;
		this.model.getDataVector().removeAllElements();

		final String[] keys = this.struct.getKeys();
		Arrays.sort(keys);

		for (final String key : keys) {
			this.model.addData(key, this.struct.get(key));
		}

		this.columsResize(this.table, 90);

		this.removeNewVarRow();
		this.addNewVarRow();

	}

	@Override
	public void saveTData(final File f) {

		final int[] selectedRows = this.table.getSelectedRows();
		final TStruct saving = new TStruct();

		if ((selectedRows.length == 0) || (selectedRows[0] > this.table.getRowCount())
				|| (selectedRows[0] > (this.struct.getKeys().length - 1))) {
			for (final String key : this.struct.getKeys()) {
				saving.put(key, this.struct.get(key));
			}
		} else {
			for (final int i : selectedRows) {

				final String key = this.model.getValueAt(i, 0).toString();
				saving.put(key, this.struct.get(key));

			}
		}

		try {
			saving.write(new FileOutputStream(f));
		} catch (final FileNotFoundException e) {
			e.printStackTrace();
		} catch (final XMLStreamException e) {
			e.printStackTrace();
		}

	}

	/**
	 * @param selection
	 *            TData
	 */
	@Override
	public void updateChangedSignal(final TData selection) {

		if (this.selectedRow <= (this.struct.getKeys().length - 1)) {
			String key = this.model.getValueAt(this.selectedRow, 0).toString();
			key = key.replaceAll(" ", "");
			this.struct.put(key, selection);
		}
	}

	/**
	 * Creats the Table Titels
	 */
	private void createTitels() {
		this.titles = new Vector<String>();
		this.titles.add(" Name ");
		this.titles.add(" Datatype ");

		this.titles.add(" Info ");
	}

	/**
	 * Creats the ToolBar
	 */
	private void createToolBar() {
		this.tb = new JToolBar();
		this.tb.setPreferredSize(new Dimension(ImageObserver.WIDTH, 25));
		this.Buttons = new Vector<JButton>();

		final JButton up = new JButton(IconEditor.ICON_UP);
		up.setActionCommand("up");
		up.setToolTipText("move one level up");
		this.Buttons.add(up);

		final JButton impor = new JButton(IconEditor.ICON_IMPORT);
		impor.setActionCommand("Import Data");
		impor.setToolTipText("import data");
		this.Buttons.add(impor);

		final JButton save = new JButton(IconEditor.ICON_SAVE);
		save.setActionCommand("Save Data");
		save.setToolTipText("save data");
		this.Buttons.add(save);

		final JButton del = new JButton(IconEditor.ICON_REMOVE);
		del.setActionCommand("Delete");
		del.setToolTipText("delete selected signal");
		this.Buttons.add(del);

		final JButton print = new JButton(IconEditor.ICON_PRINTER);
		print.setActionCommand("Print");
		print.setToolTipText("print");
		this.Buttons.add(print);

		final JButton os = new JButton(IconEditor.ICON_OpenSelection);
		os.setActionCommand("Open Selection");
		os.setToolTipText("open selection");
		this.Buttons.add(os);

		for (final JButton i : this.Buttons) {
			this.tb.add(i);
		}
		this.tb.setVisible(true);
		this.panel.add(this.tb, BorderLayout.NORTH);
		this.panel.repaint();
		this.panel.revalidate();

	}

	/**
	 * initializes the MainTable
	 */
	private void initialize() {
		this.panel = new JPanel();
		this.struct = new TStruct();
		this.panel.setLayout(new BorderLayout());
		this.createToolBar();
		this.createTitels();
		this.model = new TableModelListPanel(this.titles, 0);

		this.panel.setVisible(true);

		this.table = new JTable(this.model);
		this.table.setAutoResizeMode(0);

		this.table.setDefaultEditor(Object.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(final JTable table, final Object value,
					final boolean isSelected, final int row, final int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();

				return c;
			}
		});

		final JComboBox dataTypes = new JComboBox();

		dataTypes.addItem("Array");
		dataTypes.addItem("Boolean");
		dataTypes.addItem("Null");
		dataTypes.addItem("Numeric");
		dataTypes.addItem("Periodic");
		dataTypes.addItem("PeriodicMatrix");
		dataTypes.addItem("String");
		dataTypes.addItem("Struct");

		this.table.getColumnModel().getColumn(1).setCellEditor(new DefaultCellEditor(dataTypes));

		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.getTableHeader().setResizingAllowed(false);

		this.panel.add(new JScrollPane(this.table));

	}

	/**
	 * creats a new TData
	 * 
	 * @param datatype
	 *            Datatype
	 * @return TData
	 */
	private TData newTData(String datatype) {
		TData data = null;
		datatype = "T" + datatype;

		if (datatype.equals("TPeriodicMatrix")) {
			try {
				data = new TPeriodicMatrix(1, 1);
			} catch (final TDataException e2) {
			}
		} else {
			try {
				@SuppressWarnings("rawtypes")
				final Class cls = Class.forName("at.lcm.bu21.general.dtypes." + datatype);

				final Object o = cls.newInstance();

				data = (TData) o;

			} catch (final Exception e) {
				if (datatype.equals("TBoolean")) {
					data = new TBoolean(false);
				}
				if (datatype.equals("TString")) {
					data = new TString("");
				}

				if (datatype.equals("TNumeric")) {
					try {
						data = new TNumeric("0.0");
					} catch (final TDataException e1) {
						e1.printStackTrace();
					}
				}

			}
		}
		return data;
	}

	/**
	 * Adds the adding row
	 */
	void addNewVarRow() {

		if (this.newVarRow == false) {

			final Object[] o = new Object[3];

			o[0] = "[NAME]";
			o[1] = "Null";
			o[2] = "add a new Variable.";

			if (this.model.getColumnCount() == 1) {
				this.model.addColumn("1");
			}

			this.model.addRow(o);

			this.columsResize(this.table, 90);
			this.newVarRow = true;

		}
	}

	/**
	 * @return the model
	 */
	TableModelListPanel getModel() {
		return this.model;
	}

	/**
	 * removes the adding row
	 */
	void removeNewVarRow() {
		if (this.newVarRow == true) {
			this.model.removeRow(this.model.getRowCount() - 1);
		}
		this.newVarRow = false;
	}
}
