/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.stream.XMLStreamException;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TArrayModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TBooleanModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TDataModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TDataParentModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TNullModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TNumericModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicMatrixModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TStringModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TStructModel;
import at.lcm.bu21.general.utils.CloneException;

/**
 * This is the main class. It allows to manipulate TData types.
 * 
 * @author Johannes Kaisinger
 */
public class TDataEditor {

	/** Column_with */
	public static final int COL_WIDTH = 50;

	/**
	 * Types for the comboBox
	 * 
	 * @author Johannes Kaisinger
	 *
	 */
	private enum TDataNames {
		/** TData - TNull */
		TNull,
		/** TData - TBoolean */
		TBoolean,
		/** TData - TString */
		TString,
		/** TData - TNumeric */
		TNumeric,
		/** TData - TPeriodic */
		TPeriodic,
		/** TData - TPeriodicMatrix */
		TPeriodicMatrix,
		/** TData - TArray */
		TArray,
		/** TData - TStruct */
		TStruct
	}

	/** Model to store the data. */
	private TDataModel dataModel;

	/** Main toolBar for navigation. */
	private JToolBar tb;

	/** Current editor panel depending on the data. */
	private JPanel currentEditorPanel;

	/** This label shows the structure to the user. */
	private JLabel structureInfo;

	/** ComboBox to choose the TData-type. */
	private JComboBox<TDataNames> comboSelectTData;

	/** Listener for the comboBox. */
	private ItemListener comboSelectListener;

	/** Editor JPanel */
	private final JPanel panel;

	/** Button to go a level up in data */
	private JButton up;

	/**
	 * Creates an object
	 * 
	 * @param data
	 *            Data which should be manipulated.
	 */
	public TDataEditor(TData data) {
		TData localData = null;

		try {
			localData = data.deepClone();
		} catch (CloneException e) {
		}

		this.panel = new JPanel();
		this.panel.setLayout(new BorderLayout());

		Dimension d = new Dimension(400, 600);
		this.panel.setMinimumSize(d);

		this.createToolBar();

		this.dataModel = TDataModel.createTDataModel(localData);
		this.dataModel.setParentModel(null);

		this.setUpListener();
		this.showModel();
	}

	/**
	 * Return the manipulated TData.
	 * 
	 * @return data
	 */
	public TData getData() {
		TDataModel model;

		model = this.dataModel;
		while (model.getParentModel() != null) {
			model.getParentModel().replaceChild(model.getData());
			model = model.getParentModel();
		}

		return model.getData();
	}

	/**
	 * Show the TDataModels as Panel.
	 */
	public void showModel() {

		this.comboSelectTData.removeItemListener(this.comboSelectListener);

		if (this.currentEditorPanel != null) {
			this.panel.remove(this.currentEditorPanel);
		}

		if (this.dataModel instanceof TNullModel) {

			this.currentEditorPanel = new TNullPanel((TNullModel) this.dataModel);
			this.comboSelectTData.setSelectedItem(TDataNames.TNull);

		} else if (this.dataModel instanceof TBooleanModel) {

			this.currentEditorPanel = new TBooleanPanel((TBooleanModel) this.dataModel);
			this.comboSelectTData.setSelectedItem(TDataNames.TBoolean);

		} else if (this.dataModel instanceof TStringModel) {

			this.currentEditorPanel = new TStringPanel((TStringModel) this.dataModel);
			this.comboSelectTData.setSelectedItem(TDataNames.TString);

		} else if (this.dataModel instanceof TNumericModel) {

			this.currentEditorPanel = new TNumericPanel((TNumericModel) this.dataModel);
			this.comboSelectTData.setSelectedItem(TDataNames.TNumeric);

		} else if (this.dataModel instanceof TPeriodicModel) {

			this.currentEditorPanel = new TPeriodicPanel((TPeriodicModel) this.dataModel);
			this.comboSelectTData.setSelectedItem(TDataNames.TPeriodic);

		} else if (this.dataModel instanceof TPeriodicMatrixModel) {

			TPeriodicMatrixPanel tPeriodicMatrixPanel = new TPeriodicMatrixPanel((TPeriodicMatrixModel) this.dataModel);
			tPeriodicMatrixPanel.setEditChildListener(new EditChildListener() {

				@Override
				public void editChildOccurred(EditChildEvent evt) {

					TPeriodicMatrixModel tPeriodicMatrixModel = (TPeriodicMatrixModel) TDataEditor.this.dataModel;

					TDataEditor.this.dataModel = tPeriodicMatrixModel.getChildModel(evt.getRow(), evt.getColumn());
					TDataEditor.this.dataModel.setParentModel(tPeriodicMatrixModel);

					TDataEditor.this.showModel();

				}

			});

			this.currentEditorPanel = tPeriodicMatrixPanel;
			this.comboSelectTData.setSelectedItem(TDataNames.TPeriodicMatrix);

		} else if (this.dataModel instanceof TArrayModel) {

			TArrayPanel tArrayPanel = new TArrayPanel((TArrayModel) this.dataModel);
			tArrayPanel.setEditChildListener(new EditChildListener() {

				@Override
				public void editChildOccurred(EditChildEvent evt) {
					TArrayModel tArrayModel = (TArrayModel) TDataEditor.this.dataModel;

					TDataEditor.this.dataModel = tArrayModel.getChildModel(evt.getRow(), evt.getColumn());
					TDataEditor.this.dataModel.setParentModel(tArrayModel);

					TDataEditor.this.showModel();
				}
			});
			this.currentEditorPanel = tArrayPanel;
			this.comboSelectTData.setSelectedItem(TDataNames.TArray);

		} else if (this.dataModel instanceof TStructModel) {

			TStructPanel tStructPanel = new TStructPanel((TStructModel) this.dataModel);
			tStructPanel.setEditChildListener(new EditChildListener() {
				@Override
				public void editChildOccurred(EditChildEvent evt) {
					TStructModel tStructModel = (TStructModel) TDataEditor.this.dataModel;

					TDataEditor.this.dataModel = tStructModel.getChildModel(evt.getChildKey());
					TDataEditor.this.dataModel.setParentModel(tStructModel);

					TDataEditor.this.showModel();
				}
			});
			this.currentEditorPanel = tStructPanel;
			this.comboSelectTData.setSelectedItem(TDataNames.TStruct);

		} else {
			JOptionPane.showMessageDialog(new JDialog(), "Not Implemented yet");
		}

		TDataEditor.this.structureInfo.setText(TDataEditor.this.dataModel.getTDataStructure());
		this.up.setEnabled(TDataEditor.this.dataModel.hasParent());

		this.panel.add(this.currentEditorPanel);
		this.panel.revalidate();
		this.panel.repaint();

		this.comboSelectTData.addItemListener(this.comboSelectListener);

	}

	/**
	 * Set up the comboBox Listener.
	 */
	private void setUpListener() {
		this.comboSelectListener = new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.DESELECTED) {

					TDataEditor.this.panel.remove(TDataEditor.this.currentEditorPanel);

					TDataModel newDataModel = null;
					String value = TDataEditor.this.comboSelectTData.getSelectedItem().toString();

					switch (TDataNames.valueOf(value)) {

					case TNull:
						newDataModel = new TNullModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TDataEditor.this.currentEditorPanel = new TNullPanel((TNullModel) TDataEditor.this.dataModel);
						break;

					case TBoolean:
						newDataModel = new TBooleanModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TDataEditor.this.currentEditorPanel = new TBooleanPanel(
								(TBooleanModel) TDataEditor.this.dataModel);
						break;

					case TString:
						newDataModel = new TStringModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TDataEditor.this.currentEditorPanel = new TStringPanel(
								(TStringModel) TDataEditor.this.dataModel);
						break;

					case TNumeric:
						newDataModel = new TNumericModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TDataEditor.this.currentEditorPanel = new TNumericPanel(
								(TNumericModel) TDataEditor.this.dataModel);
						break;

					case TPeriodic:
						newDataModel = new TPeriodicModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TDataEditor.this.currentEditorPanel = new TPeriodicPanel(
								(TPeriodicModel) TDataEditor.this.dataModel);
						break;

					case TPeriodicMatrix:
						newDataModel = new TPeriodicMatrixModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());

						TDataEditor.this.dataModel = newDataModel;
						TPeriodicMatrixPanel tPeriodicMatrixPanel = new TPeriodicMatrixPanel(
								(TPeriodicMatrixModel) TDataEditor.this.dataModel);

						tPeriodicMatrixPanel.setEditChildListener(new EditChildListener() {

							@Override
							public void editChildOccurred(EditChildEvent evt) {
								TPeriodicMatrixModel tPeriodicMatrixModel = (TPeriodicMatrixModel) TDataEditor.this.dataModel;

								TDataEditor.this.dataModel = tPeriodicMatrixModel.getChildModel(evt.getRow(),
										evt.getColumn());
								TDataEditor.this.dataModel.setParentModel(tPeriodicMatrixModel);

								TDataEditor.this.showModel();

							}

						});
						TDataEditor.this.currentEditorPanel = tPeriodicMatrixPanel;
						break;

					case TArray:
						newDataModel = new TArrayModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TArrayPanel tArrayPanel = new TArrayPanel((TArrayModel) TDataEditor.this.dataModel);
						tArrayPanel.setEditChildListener(new EditChildListener() {
							@Override
							public void editChildOccurred(EditChildEvent evt) {
								TArrayModel tArrayModel = (TArrayModel) TDataEditor.this.dataModel;

								TDataEditor.this.dataModel = tArrayModel.getChildModel(evt.getRow(), evt.getColumn());
								TDataEditor.this.dataModel.setParentModel(tArrayModel);

								TDataEditor.this.showModel();
							}
						});
						TDataEditor.this.currentEditorPanel = tArrayPanel;
						break;

					case TStruct:
						newDataModel = new TStructModel();
						newDataModel.setParentModel(TDataEditor.this.dataModel.getParentModel());
						TDataEditor.this.dataModel = newDataModel;
						TStructPanel tStructPanel = new TStructPanel((TStructModel) TDataEditor.this.dataModel);
						tStructPanel.setEditChildListener(new EditChildListener() {
							@Override
							public void editChildOccurred(EditChildEvent evt) {
								TStructModel tStructModel = (TStructModel) TDataEditor.this.dataModel;

								TDataEditor.this.dataModel = tStructModel.getChildModel(evt.getChildKey());
								TDataEditor.this.dataModel.setParentModel(tStructModel);

								TDataEditor.this.showModel();
							}
						});
						TDataEditor.this.currentEditorPanel = tStructPanel;
						break;

					default:
						break;
					}

					TDataEditor.this.panel.add(TDataEditor.this.currentEditorPanel);
					TDataEditor.this.panel.revalidate();
					TDataEditor.this.panel.repaint();
				}
			}
		};
	}

	/**
	 * Create the toolBar.
	 */
	private void createToolBar() {

		this.tb = new JToolBar();

		this.up = new JButton(IconEditor.ICON_UP);
		this.up.setActionCommand("up");
		this.up.setToolTipText("move one level up");
		this.up.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				if (TDataEditor.this.dataModel.getParentModel() != null) {

					TData child = TDataEditor.this.dataModel.getData();
					TDataParentModel parentModel = TDataEditor.this.dataModel.getParentModel();
					parentModel.replaceChild(child);
					TDataEditor.this.dataModel = parentModel;

					TDataEditor.this.showModel();
				}

			}
		});
		this.tb.add(this.up);

		JButton load = new JButton(IconEditor.ICON_IMPORT);
		load.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				final JFileChooser fc = TDataEditor.this.createFileChooser();

				final int rval = fc.showOpenDialog(TDataEditor.this.panel);

				if (rval == JFileChooser.APPROVE_OPTION) {

					FileInputStream is = null;
					try {
						is = new FileInputStream(fc.getSelectedFile());
					} catch (FileNotFoundException e1) {
					}

					try {
						TData importTData = TDataEditor.this.dataModel.getData();
						importTData = TData.read(is);

						TDataParentModel parent = TDataEditor.this.dataModel.getParentModel();
						TDataEditor.this.dataModel = TDataModel.createTDataModel(importTData);
						TDataEditor.this.dataModel.setParentModel(parent);

					} catch (TDataException e1) {
					} catch (XMLStreamException e1) {
					} catch (IOException e1) {
					}

					TDataEditor.this.showModel();

				} else if (rval == JFileChooser.CANCEL_OPTION) {
				}

			}
		});
		this.tb.add(load);

		JButton save = new JButton(IconEditor.ICON_SAVE);
		save.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				final JFileChooser fc = TDataEditor.this.createFileChooser();

				final int rval = fc.showSaveDialog(TDataEditor.this.panel);

				if (rval == JFileChooser.APPROVE_OPTION) {
					File f = fc.getSelectedFile();

					if (!f.toString().endsWith(".xml")) {
						f = new File(f.getPath() + ".xml");
					}

					try {

						TDataEditor.this.dataModel.getData().write(new FileOutputStream(f));

					} catch (FileNotFoundException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					} catch (XMLStreamException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}

				} else if (rval == JFileChooser.CANCEL_OPTION) {
				}
			}
		});
		this.tb.add(save);

		this.tb.addSeparator();

		this.comboSelectTData = new JComboBox<TDataNames>(TDataNames.values());
		this.comboSelectTData.setToolTipText("choose data type");
		this.tb.add(this.comboSelectTData);

		this.structureInfo = new JLabel();

		this.panel.add(this.tb, BorderLayout.NORTH);
		this.panel.add(this.structureInfo, BorderLayout.SOUTH);

		this.panel.repaint();
		this.panel.revalidate();
	}

	/**
	 * Returns the main editor panel.
	 * 
	 * @return panel
	 */
	public JPanel getPanel() {
		return this.panel;
	}

	/**
	 * Generates a FileChooser with *.xml Filter
	 * 
	 * @return fc JFileChooser
	 */
	private JFileChooser createFileChooser() {

		final FileNameExtensionFilter filter = new FileNameExtensionFilter("xml files", "xml");
		final JFileChooser fc = new JFileChooser();
		fc.addChoosableFileFilter(filter);
		return fc;

	}
}
