/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2.TDataModel;

import java.util.LinkedList;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.TPeriodicMatrix;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Representation of a TDataModel - type
 * 
 * @author Johannes Kaisinger
 */
public abstract class TDataModel {

	/** Reference of the parent data model */
	private TDataParentModel parent;

	/**
	 * Create object with no parent data model
	 */
	public TDataModel() {
		this.parent = null;
	}

	/**
	 * Create a TDataModel depending on the TData-type
	 * 
	 * @param data
	 *            TData which is used for the model
	 * @return dataModel TDataModel of the data
	 * 
	 */
	public static TDataModel createTDataModel(TData data) {
		TDataModel dataModel = null;

		if (data.getClass() == TNull.class) {
			dataModel = new TNullModel((TNull) data);
		} else if (data.getClass() == TBoolean.class) {
			dataModel = new TBooleanModel((TBoolean) data);
		} else if (data.getClass() == TString.class) {
			dataModel = new TStringModel((TString) data);
		} else if (data.getClass() == TNumeric.class) {
			dataModel = new TNumericModel((TNumeric) data);
		} else if (data.getClass() == TPeriodic.class) {
			dataModel = new TPeriodicModel((TPeriodic) data);
		} else if (data.getClass() == TPeriodicMatrix.class) {
			dataModel = new TPeriodicMatrixModel((TPeriodicMatrix) data);
		} else if (data.getClass() == TArray.class) {
			dataModel = new TArrayModel((TArray) data);
		} else if (data.getClass() == TStruct.class) {
			dataModel = new TStructModel((TStruct) data);
		}

		return dataModel;
	}

	/**
	 * Create object with given data (TDataParentModel)
	 * 
	 * @param parent
	 *            Reference of parent data model (TDataParentModel)
	 */
	public TDataModel(TDataParentModel parent) {
		this.parent = parent;
	}

	/**
	 * Set parent data model (TDataParentModel)
	 * 
	 * @param parent
	 *            Reference of parent data model (TDataParentModel)
	 */
	public void setParentModel(TDataParentModel parent) {
		this.parent = parent;
	}

	/**
	 * Get parent data model (TDataParentModel)
	 * 
	 * @return parent Reference of parent data model (TDataParentModel)
	 */
	public TDataParentModel getParentModel() {
		return this.parent;
	}

	/**
	 * Get all data (TData)
	 * 
	 * @return Data Reference of data (TData)
	 */
	public abstract TData getData();

	/**
	 * Returns the actual position in the main data
	 * 
	 * @return String of structure
	 */
	public String getTDataStructure() {
		String fullStructure = "";

		TDataModel model = this;

		LinkedList<String> structure = new LinkedList<String>();

		while (model.getParentModel() != null) {
			model = model.getParentModel();
			structure.addLast(((TDataParentModel) model).getChildName());
		}

		structure.addLast("data");

		for (int i = structure.size() - 1; i >= 0; i--) {
			fullStructure += "." + structure.get(i);
		}

		return fullStructure;
	}

	/**
	 * Checks if the model has a parent model
	 * 
	 * @return hasParent
	 */
	public boolean hasParent() {
		if (this.parent != null) {
			return true;
		} else {
			return false;
		}
	}
}
