/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1027 $
 * $Date: 2020-07-07 11:31:18 +0200 (Tue, 07 Jul 2020) $
 */
package at.lcm.bu21.general.dtypes.preview;

import org.apache.commons.math3.exception.MathParseException;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.plot.Plot2Bar;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.PlotException;
import at.lcm.bu21.general.utils.CloneException;

/**
 * TPeriodic preview methods and EPS export
 * 
 * @author gweidenholzer
 * 
 */
public class TPeriodicPreview {

	/**
	 * Get a preview-plot
	 * 
	 * @param signal
	 *            Signal to display
	 * @param title
	 *            Title of preview
	 * @param nPeriods
	 *            Number of periods to display
	 * @param showFundamental
	 *            display fundamental oscillation if true
	 * @return panel with preview
	 */
	public static Plot2Line previewWave(TPeriodic signal, String title, int nPeriods, boolean showFundamental) {
		Plot2Line plot;
		final int N = 500;

		plot = new Plot2Line();
		plot.setTitle(title);
		plot.setXLabel("Angle / degree");
		plot.setYLabel("Magnitude");
		try {
			plot.addSignal("Signal", signal, N);
		} catch (PlotException e) {
		}
		if (showFundamental) {
			try {
				plot.addSignal("Fundamental", signal.select(new int[] { 1 }), N);
			} catch (PlotException e) {
			}
		}

		return plot;
	}

	/**
	 * Get a spectrum-plot
	 * 
	 * @param signal
	 *            Signal to display
	 * @param title
	 *            Title of preview
	 * @return Plot with spectrum preview
	 */
	public static Plot2Bar previewSpectrum(TPeriodic signal, String title) {
		return TPeriodicPreview.previewSpectrum(signal, title, signal.getMaxOrder(), false);
	}

	/**
	 * Get a spectrum-plot
	 * 
	 * @param signal
	 *            Signal to display
	 * @param title
	 *            Title of preview
	 * @param maxorder
	 *            highest order
	 * @param rms
	 *            true: rms, false: magnitude
	 * @return Plot with spectrum preview
	 */
	public static Plot2Bar previewSpectrum(TPeriodic signal, String title, int maxorder, boolean rms) {

		Plot2Bar plot;
		double[] orders, values;
		double mag1;

		int n = signal.getMaxOrder();
		if (n > maxorder) {
			n = maxorder;
		}

		orders = new double[n + 1];
		values = new double[n + 1];
		orders[0] = 0.0;
		values[0] = signal.mean();

		for (int k = 1; k <= n; k++) {
			orders[k] = k;

			if (rms) {
				values[k] = signal.magnitude(k) / Math.sqrt(2.);
			} else {
				values[k] = signal.magnitude(k);
			}
		}

		plot = new Plot2Bar();
		plot.setTitle(title);
		plot.setXLabel("Order");

		if (rms) {
			plot.setYLabel("RMS");

		} else {
			plot.setYLabel("Magnitude");
		}

		try {
			plot.addSignal("Spectrum", orders, values);
		} catch (MathParseException e) {
		} catch (PlotException e) {
		} catch (TDataException e) {
		} catch (CloneException e) {
		}

		mag1 = signal.magnitude(1);
		if (mag1 > 0.0) {
			plot.addYAxis("Magnitude / %", 0.0, 100.0 / mag1);
		}

		return plot;
	}

	/**
	 * Get a preview-panel of 2d parametric curve x(phi), y(phi) with x(phi)
	 * being the object itself
	 * 
	 * @param x
	 *            x-signal
	 * 
	 * @param y
	 *            y-signal
	 * @param title
	 *            Title of preview
	 * @return panel with 2d curve preview
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public static Plot2Line preview2d(TPeriodic x, TPeriodic y, String title) throws CloneException {
		Plot2Line plot;

		final int N = 500;

		plot = new Plot2Line();
		plot.setTitle(title);
		plot.setXLabel("x");
		plot.setYLabel("y");
		try {
			plot.addSignal("Signal", x, y, N);
		} catch (MathParseException e) {
		} catch (PlotException e) {
		} catch (TDataException e) {
		}

		return plot;
	}

}
