/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1022 $
 * $Date: 2020-06-03 14:13:51 +0200 (Wed, 03 Jun 2020) $
 */
package at.lcm.bu21.general.dtypes.unit;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

/**
 * Scan unit and return Tokens (lexical analyze)
 *
 * @author Werner Koppelsttter
 */
public class Scanner {

	/** End of file character */
	static final char eofCh = (char) -1;

	/** Input as reader */
	private Reader in;
	/** Next character */
	private char ch;
	/** Row of next character */
	int line;
	/** Column of next character */
	int col;

	/**
	 * Initialize scanner with unit-string
	 * 
	 * @param unit
	 *            Expression to scan
	 */
	public Scanner(String unit) {
		this(new StringReader(unit));
	}

	/**
	 * Initialize scanner with reader
	 * 
	 * @param reader
	 *            Reader with unit
	 */
	public Scanner(Reader reader) {
		this.in = reader;
		this.line = 1;
		this.col = 0;
		this.nextCh();
	}

	/**
	 * Read next character
	 */
	private void nextCh() {
		try {
			this.ch = (char) this.in.read();
			this.col++;
			if (this.ch == '\n') {
				this.line++;
				this.col = 0;
			}
		} catch (IOException ex) {
			this.ch = Scanner.eofCh;
		}
	}

	/**
	 * Read next token
	 * 
	 * @return Scanned token
	 * @throws ScannerException
	 */
	public Token next() throws ScannerException {
		Token tok;
		while ((this.ch <= ' ') || (this.ch == '.') || (this.ch == '*') || (this.ch == '') || (this.ch == '^')) {
			this.nextCh();
		}

		tok = new Token();
		tok.line = this.line;
		tok.col = this.col;

		switch (this.ch) {
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'g':
		case 'h':
		case 'i':
		case 'j':
		case 'k':
		case 'l':
		case 'm':
		case 'n':
		case 'o':
		case 'p':
		case 'q':
		case 'r':
		case 's':
		case 't':
		case 'u':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z':
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'G':
		case 'H':
		case 'I':
		case 'J':
		case 'K':
		case 'L':
		case 'M':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'S':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '':
		case '%':
			this.readName(tok);
			break;
		case '-':
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			this.readNumber(tok);
			break;
		case '/':
			this.nextCh();
			tok.kind = TokenKind.SLASH;
			break;
		case '(':
			this.nextCh();
			tok.kind = TokenKind.LPAR;
			break;
		case ')':
			this.nextCh();
			tok.kind = TokenKind.RPAR;
			break;
		case '?':
			this.nextCh();
			tok.kind = TokenKind.QUESTION;
			break;
		case eofCh:
			tok.kind = TokenKind.EOF;
			break;
		default:
			tok.str = "" + this.ch;
			this.nextCh();
			tok.kind = TokenKind.NONE;
			throw new ScannerException(this, "Unexpected sign " + this.ch);
		}

		return tok;
	}

	/**
	 * Read names and keywords
	 * 
	 * @param tok
	 *            Token to insert name
	 */
	private void readName(Token tok) {
		StringBuilder buf = new StringBuilder();

		buf.append(this.ch);
		this.nextCh();
		while ((('a' <= this.ch) && (this.ch <= 'z')) || (('A' <= this.ch) && (this.ch <= 'Z')) || (this.ch == '%')) {
			buf.append(this.ch);
			this.nextCh();
		}

		tok.str = buf.toString();
		tok.kind = TokenKind.ELEMENT;
	}

	/**
	 * Read numbers (real or imaginary)
	 * 
	 * @param tok
	 *            Token to insert number
	 * @throws ScannerException
	 */
	private void readNumber(Token tok) throws ScannerException {
		StringBuilder buf = new StringBuilder();

		buf.append(this.ch);
		this.nextCh();
		while (('0' <= this.ch) && (this.ch <= '9')) {
			buf.append(this.ch);
			this.nextCh();
		}

		try {
			tok.kind = TokenKind.NUMBER;
			tok.val = Integer.parseInt(buf.toString());
		} catch (Exception e) {
			tok.str = buf.toString();
			tok.kind = TokenKind.NONE;
			throw new ScannerException(this, "Unable to parse number '" + tok.str + "'");
		}
	}
}
