/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.math.autodiff;

/**
 * Basic function of all defined functions
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public abstract class Func {

	/** List of arguments of this function */
	private Func[] args;

	/**
	 * Create function with variable count of arguments
	 * 
	 * @param args
	 *            Variable count of arguments
	 */
	protected Func(Func... args) {
		this.args = args;
	}

	/**
	 * Get argument of index
	 * 
	 * @param index
	 *            Index of argument
	 * @return Argument of index
	 * @throws Exception
	 */
	protected Func getArg(int index) throws Exception {
		if (index >= this.args.length) {
			throw new Exception("Wrong number of function arguments!");
		}
		return this.args[index];
	}

	/**
	 * Get number of arguments
	 * 
	 * @return Number of arguments
	 */
	protected int getNumArgs() {
		return this.args.length;
	}

	/**
	 * Evaluate function and return value
	 * 
	 * @return value Value of evaluated function
	 * @throws Exception
	 *             Eval exception
	 */
	public abstract double value() throws Exception;

	/**
	 * Derivation of function
	 * 
	 * @param arg
	 *            Argument for derivative
	 * @return Derivative function
	 * @throws Exception
	 *             Derivative exception
	 */
	protected Func derivativeArg(Func arg) throws Exception {
		throw new Exception("Derivative not defined");
	}

	/**
	 * Chain rule for derivation
	 * 
	 * @param var
	 *            Variable for derivative
	 * @return Derivative function
	 * @throws Exception
	 *             Derivative exception
	 */
	public Func derivative(Variable var) throws Exception {
		Func[] prods = new Func[this.getNumArgs()];

		for (int k = 0; k < this.getNumArgs(); k++) {
			prods[k] = new Product(this.derivativeArg(this.getArg(k)),
					this.getArg(k).derivative(var));
		}

		return new Sum(prods);

	}

	@Override
	public String toString() {
		String str;

		str = this.getClass().getSimpleName() + "(";
		try {
			str += this.getArg(0).toString();
		} catch (Exception e) {
			str += "'?'";
		}
		for (int k = 1; k < this.getNumArgs(); k++) {
			str += ",";
			try {
				str += this.getArg(k).toString();
			} catch (Exception e) {
				str += "'?'";
			}
		}
		str += ")";

		return str;
	}

	// /**
	// * Return apache-commons-math compatible function
	// *
	// * @param vars
	// * Variable count of variable for differential function
	// * @return Apache-commons-math compatible function
	// */
	// public DifferentiableMultivariateRealFunction
	// getDifferentiableMultivariateFunction(
	// Variable... vars) {
	// return new DifferentiableMultivariateFunction(this, vars);
	// }
	//
	// /**
	// * @param vars
	// * @return function
	// */
	// public DifferentiableMultivariateRealFunction
	// getDifferentiableMultivariateFunction2(
	// Variable... vars) {
	// return new DifferentiableMultivariateFunction2(this, vars);
	// }
	//
	//
	//
	// /**
	// * Return apache-commons-math compatible function
	// *
	// * @param var
	// * Variabl for differential function
	// * @return Apache-commons-math compatible function
	// */
	// public DifferentiableUnivariateRealFunction getDifferentiableFunction(
	// Variable var) {
	// return new DifferentiableFunction(this, var);
	// }

}
