/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1027 $
 * $Date: 2020-07-07 11:31:18 +0200 (Tue, 07 Jul 2020) $
 */
package at.lcm.bu21.general.plot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;

import javax.swing.SwingUtilities;

import org.freehep.graphicsio.AbstractVectorGraphicsIO;
import org.freehep.graphicsio.pdf.PDFGraphics2D;
import org.freehep.graphicsio.ps.PSGraphics2D;
import org.freehep.graphicsio.svg.SVGGraphics2D;
import org.jfree.chart.ChartMouseEvent;
import org.jfree.chart.ChartMouseListener;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.StandardChartTheme;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.XYItemEntity;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.plot.DatasetRenderingOrder;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.plot.XYPolarPlot;
import org.jfree.chart.renderer.ColorPaintScale;
import org.jfree.chart.renderer.xy.DefaultXYItemRenderer;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYBlockRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYSplineRenderer;
import org.jfree.chart.renderer.xy.XYStepAreaRenderer;
import org.jfree.chart.renderer.xy.XYStepRenderer;
import org.jfree.chart.title.PaintScaleLegend;
import org.jfree.data.xy.DefaultXYZDataset;
import org.jfree.data.xy.XYBarDataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.RectangleInsets;

import at.lcm.bu21.general.utils.MathUtils;

/**
 * Base type of TData plotting classes
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public abstract class Plot2D extends Plot {

	/** Chart for plotting signals */
	protected JFreeChart chart;

	/** Plot for plotting signals */
	protected XYPlot plot;

	/** ChartMouseListener (for Marking and Zooming) */
	private ChartMouseListener chartMouseListener;

	/** Key Listener */
	private KeyListener keyListener;

	/** MouseListener (for Marking and Zooming) */
	private MouseListener mouseListener;

	/** MouseMotionListener (for Marking and Zooming) */
	private MouseMotionListener mouseMotionListener;

	/** contains Mouse Pressed x/y position */
	private Point2D mousePressedPositionChart;

	/** Point, needed for marking Points (Mouse Button Pressed) */
	private Point mousePressedPositionOnScreen;

	/** contains Mouse Released x/y position */
	private Point2D mouseReleasedPositionChart;

	/** Point, needed for marking Points (Mouse Button Released) */
	private Point mouseReleasedPositionOnScreen;

	/** is true, id the XBounds has been set */
	protected boolean xLimitSet = false;

	/** min. X Axis Limit */
	double xLimitMin;

	/** max. X Axis Limit */
	double xLimitMax;

	/** min. X Axis boundary */
	double xBoundMin;

	/** max. X Axis boundary */
	double xBoundMax;

	/** is true, id the YBounds has been set */
	protected boolean yLimitSet = false;

	/** min. Y Axis Limit */
	protected double yLimitMin;

	/** max. Y Axis Limit */
	protected double yLimitMax;

	/** min. Y Axis boundary */
	protected double yBoundMin;

	/** max. Y Axis boundary */
	protected double yBoundMax;

	/** is true, id the ZBounds has been set */
	protected boolean zLimitSet = false;

	/** min. Z Axis Limit */
	protected double zLimitMin;

	/** max. Z Axis Limit */
	protected double zLimitMax;

	/** Panel for plotting signals */
	protected ChartPanel chartPanel;

	/** Y-axis show zero */
	private boolean yAxisShowZero = true;

	/** Color transparency */
	private float colorAlpha = 1.0f;

	/** Shift and scale factor for domain-axis (x) */
	protected ArrayList<double[]> domainAxisShiftScale;

	/** Shift and scale factor for range-axis (y) */
	protected ArrayList<double[]> rangeAxisShiftScale;

	/**
	 * Create empty object
	 */
	protected Plot2D() {
		super();

		this.domainAxisShiftScale = new ArrayList<double[]>();
		this.domainAxisShiftScale.add(new double[] { 0.0, 1.0 });

		this.rangeAxisShiftScale = new ArrayList<double[]>();
		this.rangeAxisShiftScale.add(new double[] { 0.0, 1.0 });

		this.initListener();

		this.mousePressedPositionOnScreen = new Point();
		this.mouseReleasedPositionOnScreen = new Point();
		this.mousePressedPositionChart = new Point2D.Double();
		this.mouseReleasedPositionChart = new Point2D.Double();
		this.getPlotPanel().addMouseMotionListener(this.mouseMotionListener);
		this.getPlotPanel().addKeyListener(this.keyListener);

		this.plot = new XYPolarPlot(null, new NumberAxis(this.getXLabel()), new NumberAxis(this.getYLabel()),
				new DefaultXYItemRenderer());
		this.plot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);
		this.chart = new JFreeChart(this.getTitle(), JFreeChart.DEFAULT_TITLE_FONT, this.plot, true);
		this.chart.getLegend().visible = false;

		// this.chartPanel = new ChartPanel(this.chart);
		this.chartPanel = new ChartPanel(this.chart, ChartPanel.DEFAULT_WIDTH, ChartPanel.DEFAULT_HEIGHT,
				ChartPanel.DEFAULT_MINIMUM_DRAW_WIDTH, ChartPanel.DEFAULT_MINIMUM_DRAW_HEIGHT,
				ChartPanel.DEFAULT_MAXIMUM_DRAW_WIDTH, ChartPanel.DEFAULT_MAXIMUM_DRAW_HEIGHT,
				ChartPanel.DEFAULT_BUFFER_USED, false, // properties
				false, // copy
				false, // save
				false, // print
				false, // zoom
				true // tooltips
		);

		this.chartPanel.setLayout(new BorderLayout());
		this.chartPanel.addChartMouseListener(this.chartMouseListener);
		this.chartPanel.setFont(new Font("Helvetica", Font.PLAIN, 10));
		StandardChartTheme ct = (StandardChartTheme) org.jfree.chart.StandardChartTheme.createJFreeTheme();

		ct.setRegularFont(new Font("Helvetica", Font.PLAIN, 12));
		ct.setExtraLargeFont(new Font("Helvetica", Font.PLAIN, 20));
		ct.setSmallFont(new Font("Helvetica", Font.PLAIN, 12));
		ct.setLargeFont(new Font("Helvetica", Font.PLAIN, 12));

		ct.setChartBackgroundPaint(Color.WHITE);
		ct.apply(this.chart);

		this.chartPanel.setAutoscrolls(true);
		this.chartPanel.setMouseWheelEnabled(true);

		this.chartPanel.setMouseZoomable(false);

		this.chartPanel.setDomainZoomable(false);
		this.chartPanel.setRangeZoomable(false);

		this.getPlotPanel().add(this.chartPanel, BorderLayout.CENTER);
		this.chart.setAntiAlias(true);
		this.chartPanel.addMouseListener(this.mouseListener);
		this.chartPanel.addMouseMotionListener(this.mouseMotionListener);

		this.plot.getDomainAxis().setAutoRange(true);
		this.plot.getRangeAxis().setAutoRange(true);
		this.plot.setDomainGridlinePaint(Color.DARK_GRAY);
		this.plot.setRangeGridlinePaint(Color.DARK_GRAY);

		this.plot.setBackgroundPaint(Color.WHITE);
		this.plot.setAxisOffset(new RectangleInsets(0, 0, 0, 0));

		this.reset();

		this.getChartPanel().setPopupMenu(this.popup.getPopupMenu());

	}

	/**
	 * Sets the chart title
	 * 
	 * @param title
	 *            chart title
	 */
	@Override
	public void setTitle(String title) {
		super.setTitle(title);
		if (this.chart != null) {
			this.chart.setTitle(title);
		}

	}

	/**
	 * Returns the Legend visibility
	 * 
	 * @return boolean
	 */
	public boolean getLegend() {

		return this.chart.getLegend().visible;
	}

	// /**
	// * Returns selectet x and y values
	// *
	// * @return selectet Values
	// */
	// public double[][] getSelectetValues() {
	// //
	// // if (this.markerXYSeries != null) {
	// // return this.markerXYSeries.toArray();
	// // } else {
	// return null;
	// // }
	// //
	// }

	@Override
	public void reset() {
		this.redraw();
		this.zoom();
	}

	/**
	 * sets the DataCursor Visibility
	 * 
	 * @param trace
	 *            DataCursor visibility
	 */
	public void setDataCursor(boolean trace) {
		this.chartPanel.setVerticalAxisTrace(trace);
		this.chartPanel.setHorizontalAxisTrace(trace);

	}

	/**
	 * Turns the Legend on/off
	 * 
	 * @param b
	 *            Legend on/off
	 */
	public void setLegend(boolean b) {
		try {
			this.chart.getLegend().setVisible(b);
		} catch (Exception e) {
		}
		this.redraw();
	}

	/**
	 * Sets the x limit
	 * 
	 * @param min
	 *            minimum x value
	 * @param max
	 *            maximal x value
	 */
	public void setXLimit(double min, double max) {
		this.xLimitMin = min;
		this.xLimitMax = max;
		this.xLimitSet = true;
		this.xBoundMin = this.xLimitMin;
		this.xBoundMax = this.xLimitMax;
		this.updateBounds();
	}

	/**
	 * Clear the x limit
	 */
	public void clearXLimit() {
		this.xLimitSet = false;
		this.zoom();
	}

	/**
	 * Sets the y Limit
	 * 
	 * @param min
	 *            minimum y value
	 * @param max
	 *            maximal y value
	 */
	public void setYLimit(double min, double max) {
		this.yLimitMin = min;
		this.yLimitMax = max;
		this.yLimitSet = true;
		this.yBoundMin = this.yLimitMin;
		this.yBoundMax = this.yLimitMax;
		this.updateBounds();
	}

	/**
	 * Clear the y limit
	 */
	public void clearYLimit() {
		this.yLimitSet = false;
		this.zoom();
	}

	/**
	 * Sets the z limit
	 * 
	 * @param min
	 *            minimum z value
	 * @param max
	 *            maximal z value
	 */
	public void setZLimit(double min, double max) {
		this.zLimitMin = min;
		this.zLimitMax = max;
		this.zLimitSet = true;
		this.getPlotPanel().repaint();
	}

	/**
	 * Clear the x limit
	 */
	public void clearZLimit() {
		this.zLimitSet = false;
		this.getPlotPanel().repaint();
	}

	/**
	 * Initializes -Mouse - MouseMotion and ChartMouseListener
	 */
	private void initListener() {

		this.keyListener = new KeyAdapter() {

			@Override
			public void keyTyped(KeyEvent event) {

				switch (event.getKeyChar()) {
				case '':
					Plot2D.this.reset();
				}
			}
		};

		this.mouseMotionListener = new MouseMotionAdapter() {

			/**
			 * Needed for finding out the actual mouse position (for Marking
			 * helper)
			 * 
			 * @param event
			 *            MouseEvent
			 */
			@Override
			public void mouseDragged(MouseEvent event) {
				int x1, x2, y1, y2, tmp;

				if (SwingUtilities.isLeftMouseButton(event)) {
					x1 = Plot2D.this.mousePressedPositionOnScreen.x;
					y1 = Plot2D.this.mousePressedPositionOnScreen.y;
					x2 = event.getX();
					y2 = event.getY();

					if (x1 > x2) {
						tmp = x1;
						x1 = x2;
						x2 = tmp;
					}
					if (y1 > y2) {
						tmp = y1;
						y1 = y2;
						y2 = tmp;
					}

					Plot2D.this.getPlotPanel().rect = new Rectangle(x1, y1, x2 - x1, y2 - y1);
					Plot2D.this.getPlotPanel().repaint();
				}
			}
		};

		this.chartMouseListener = new ChartMouseListener() {

			/**
			 * triggered by ChartMouse Event, needed to select a point
			 * 
			 * @param chartEvent
			 *            ChartMouseEvent
			 */
			@Override
			public void chartMouseClicked(ChartMouseEvent chartEvent) {
				ChartEntity entity;
				int len;
				LinkPlot link;
				MouseEvent event;
				boolean add;
				Signal sig;
				Point2D.Double pBottomLeft, pTopRight;

				if (SwingUtilities.isLeftMouseButton(chartEvent.getTrigger())) {
					entity = chartEvent.getEntity();
					if ((entity != null) && (entity instanceof XYItemEntity)) {
						event = chartEvent.getTrigger();

						link = Plot2D.this.getLink();
						len = Plot2D.this.getNumSignals();
						if (!event.isControlDown()) {
							link.clearSelect();
						}

						pBottomLeft = Plot2D.this.transltePointToChart(event.getX() - 2, event.getY() + 2);
						pTopRight = Plot2D.this.transltePointToChart(event.getX() + 2, event.getY() - 2);
						add = !(event.isControlDown() && event.isShiftDown());
						for (int k = 0; k < len; k++) {
							sig = Plot2D.this.getSignal(k);
							if ((sig != null) && sig.isEnable()) {
								if (add) {
									Plot2D.this.select(k, pBottomLeft.getX(), pTopRight.getX(), pBottomLeft.getY(),
											pTopRight.getY(), true);
								} else {
									Plot2D.this.select(k, pBottomLeft.getX(), pTopRight.getX(), pBottomLeft.getY(),
											pTopRight.getY(), false);
								}
							}
						}
						link.redraw();
					}
				}
			}

			@Override
			public void chartMouseMoved(ChartMouseEvent charEvent) {
			}
		};

		this.mouseListener = new MouseAdapter() {

			/**
			 * Triggered by a MouseEvent, needed to select more than one points
			 */
			@Override
			public void mouseReleased(MouseEvent event) {

				LinkPlot link;
				int len;
				boolean add;
				Signal sig;
				double xp, xr, yp, yr;

				if (SwingUtilities.isLeftMouseButton(event)) {
					Plot2D.this.getPlotPanel().rect = null;
					Plot2D.this.mouseReleasedPositionOnScreen.setLocation(event.getX(), event.getY());
					Plot2D.this.mouseReleasedPositionChart.setLocation(
							Plot2D.this.transltePointToChart(event.getX(), event.getY()).getX(),
							Plot2D.this.transltePointToChart(event.getX(), event.getY()).getY());
					xp = Plot2D.this.mousePressedPositionChart.getX();
					xr = Plot2D.this.mouseReleasedPositionChart.getX();
					yp = Plot2D.this.mousePressedPositionChart.getY();
					yr = Plot2D.this.mouseReleasedPositionChart.getY();

					if (((xp < xr) && (yp < yr)) || ((xp > xr) && (yp > yr))) {
						// mark
						link = Plot2D.this.getLink();
						len = Plot2D.this.getNumSignals();
						if (!event.isControlDown()) {
							link.clearSelect();
						}
						add = !(event.isControlDown() && event.isShiftDown());
						for (int k = 0; k < len; k++) {
							sig = Plot2D.this.getSignal(k);
							if ((sig != null) && sig.isEnable()) {
								if (add) {
									Plot2D.this.select(k, Math.min(xp, xr), Math.max(xp, xr), Math.min(yr, yp),
											Math.max(yr, yp), true);
								} else {
									Plot2D.this.select(k, Math.min(xp, xr), Math.max(xp, xr), Math.min(yr, yp),
											Math.max(yr, yp), false);
								}
							}
						}
						link.redraw();
					} else if ((xp < xr) && (yp > yr)) {
						// zoom in
						Plot2D.this.xBoundMin = xp;
						Plot2D.this.xBoundMax = xr;
						Plot2D.this.yBoundMin = yr;
						Plot2D.this.yBoundMax = yp;
						Plot2D.this.updateBounds();
					} else if ((xp > xr) && (yp < yr)) {
						// zoom out
						Plot2D.this.xBoundMin = Plot2D.this.xLimitMin;
						Plot2D.this.xBoundMax = Plot2D.this.xLimitMax;
						Plot2D.this.yBoundMin = Plot2D.this.yLimitMin;
						Plot2D.this.yBoundMax = Plot2D.this.yLimitMax;
						Plot2D.this.updateBounds();
					}
				}
			}

			/**
			 * Triggered by a MouseEvent, needed to select more than one points
			 */

			@Override
			public void mousePressed(MouseEvent event) {
				if (SwingUtilities.isLeftMouseButton(event)) {
					Plot2D.this.getPlotPanel().requestFocus();
					Plot2D.this.mousePressedPositionOnScreen.setLocation(event.getX(), event.getY());
					Plot2D.this.mousePressedPositionChart.setLocation(
							Plot2D.this.transltePointToChart(event.getX(), event.getY()).getX(),
							Plot2D.this.transltePointToChart(event.getX(), event.getY()).getY());
				}
			}
		};
	}

	@Override
	void repaint() {

		this.getPlotPanel().repaint();
		this.getPlotPanel().revalidate();
		if (this.chartPanel != null) {
			this.chartPanel.repaint();
			this.chartPanel.revalidate();
		}
	}

	/**
	 * Updates the boundaries
	 */
	void updateBounds() {
		double diff, min, max;
		ValueAxis axis;

		// scale x-limits for display marker (exception for value 0.0)
		min = this.xBoundMin;
		max = this.xBoundMax;
		if (!this.xLimitSet) {
			diff = Math.abs(max - min);
			min += Math.abs(min) < (diff * 1e-5) ? -max * 0.02 : -diff * 0.02;
			max += Math.abs(max) < (diff * 1e-5) ? -min * 0.02 : +diff * 0.02;
		}
		for (int k = 0; k < this.plot.getDomainAxisCount(); k++) {
			if (k < this.domainAxisShiftScale.size()) {
				axis = this.plot.getDomainAxis(k);
				axis.setLowerBound((min + this.domainAxisShiftScale.get(k)[0]) * this.domainAxisShiftScale.get(k)[1]);
				axis.setUpperBound((max + this.domainAxisShiftScale.get(k)[0]) * this.domainAxisShiftScale.get(k)[1]);
			}
		}

		// scale y-limits for display marker (exception for value 0.0)
		min = this.yBoundMin;
		max = this.yBoundMax;
		if (!this.yLimitSet) {
			diff = Math.abs(max - min);
			min += Math.abs(min) < (diff * 1e-5) ? -max * 0.02 : -diff * 0.02;
			max += Math.abs(max) < (diff * 1e-5) ? -min * 0.02 : +diff * 0.02;
		}
		for (int k = 0; k < this.plot.getRangeAxisCount(); k++) {
			if (k < this.rangeAxisShiftScale.size()) {
				axis = this.plot.getRangeAxis(k);
				axis.setLowerBound((min + this.rangeAxisShiftScale.get(k)[0]) * this.rangeAxisShiftScale.get(k)[1]);
				axis.setUpperBound((max + this.rangeAxisShiftScale.get(k)[0]) * this.rangeAxisShiftScale.get(k)[1]);
			}
		}

		this.getPlotPanel().repaint();
	}

	/**
	 * This Methode zooms the whole plot to the right perspective
	 */
	public void zoom() {
		final int min = 0;
		final int max = 1;
		int sigs;
		double diff;

		if (!this.xLimitSet) {
			double[] limitsX = new double[] { 0.0, 1.0 };

			sigs = this.getNumSignals();
			if (this.getNumSignals() > 0) {
				double[] minX, maxX;

				minX = new double[sigs];
				maxX = new double[sigs];

				for (int k = 0; k < sigs; k++) {
					Signal sig = this.getSignal(k);
					minX[k] = sig.getMinX();
					maxX[k] = sig.getMaxX();
				}

				limitsX[min] = MathUtils.min(minX);
				limitsX[max] = MathUtils.max(maxX);
			}

			// limitsX = MathUtils.roundLimits(limitsX);

			// scale limits for display marker (exception for value 0.0)
			// x-limits
			diff = Math.abs(limitsX[max] - limitsX[min]);
			if (diff == 0.0) {
				diff = limitsX[min] * 0.04;
				if (diff < 0.0) {
					diff *= -1.0;
				}
				limitsX[min] -= 0.02 * diff;
				limitsX[max] += 0.02 * diff;
			}

			this.xLimitMin = limitsX[min];
			this.xLimitMax = limitsX[max];
		}
		this.xBoundMin = this.xLimitMin;
		this.xBoundMax = this.xLimitMax;

		if (!this.yLimitSet) {
			double[] limitsY = new double[] { 0.0, 1.0 };

			sigs = this.getNumSignals();
			if (this.getNumSignals() > 0) {
				double[] minY, maxY;

				minY = new double[sigs];
				maxY = new double[sigs];

				for (int k = 0; k < sigs; k++) {
					Signal sig = this.getSignal(k);
					minY[k] = sig.getMinY();
					maxY[k] = sig.getMaxY();
				}

				limitsY[min] = MathUtils.min(minY);
				limitsY[max] = MathUtils.max(maxY);
			}

			if (this.getYAxisShowZero()) {
				if (limitsY[0] > 0) {
					limitsY[0] = 0.0;
				}
				if (limitsY[1] < 0) {
					limitsY[1] = 0.0;
				}
			}

			// limitsY = MathUtils.roundLimits(limitsY);

			// y-limits
			diff = Math.abs(limitsY[max] - limitsY[min]);
			if (diff == 0.0) {
				diff = limitsY[min] * 0.04;
				if (diff < 0.0) {
					diff *= -1.0;
				}
				limitsY[min] -= 0.02 * diff;
				limitsY[max] += 0.02 * diff;
			}

			this.yLimitMin = limitsY[min];
			this.yLimitMax = limitsY[max];
		}
		this.yBoundMin = this.yLimitMin;
		this.yBoundMax = this.yLimitMax;

		this.updateBounds();
	}

	/**
	 * translates panel cooards. into valid data coards.
	 * 
	 * @param i
	 *            x (in pixel)
	 * @param j
	 *            y (in pixel)
	 * @return translatet Point
	 */
	protected Point2D.Double transltePointToChart(int i, int j) {

		Point2D point2d = this.chartPanel.translateScreenToJava2D(new Point(i, j));
		XYPlot xyplot = this.plot;
		ChartRenderingInfo chartrenderinginfo = this.chartPanel.getChartRenderingInfo();

		java.awt.geom.Rectangle2D rectangle2d = chartrenderinginfo.getPlotInfo().getDataArea();

		ValueAxis valueaxis = xyplot.getDomainAxis();
		org.jfree.ui.RectangleEdge rectangleedge = xyplot.getDomainAxisEdge();

		ValueAxis valueaxis1 = xyplot.getRangeAxis();
		org.jfree.ui.RectangleEdge rectangleedge1 = xyplot.getRangeAxisEdge();

		double x = valueaxis.java2DToValue(point2d.getX(), rectangle2d, rectangleedge);
		double y = valueaxis1.java2DToValue(point2d.getY(), rectangle2d, rectangleedge1);

		return new Point2D.Double(x, y);
	}

	/**
	 * Select point of signal within e rectangle
	 * 
	 * @param signal
	 *            Signal to select
	 * @param xmin
	 *            Min of x values
	 * @param xmax
	 *            Max of x values
	 * @param ymin
	 *            Min of y values
	 * @param ymax
	 *            Max of y value
	 * @param select
	 *            Select or deselect
	 */
	void select(int signal, double xmin, double xmax, double ymin, double ymax, boolean select) {
		double[] x, y;
		LinkPlot link;
		Signal sig;

		link = this.getLink();
		sig = this.getSignal(signal);
		if (sig != null) {
			x = sig.getXRef();
			y = sig.getYRef();

			for (int k = 0; k < x.length; k++) {
				if ((xmin <= x[k]) && (x[k] <= xmax) && (ymin <= y[k]) && (y[k] <= ymax)) {
					link.select(signal, k, select);
				}
			}
		}
	}

	@Override
	void redraw() {
		int sigs;

		// draw signal
		sigs = this.getNumSignals();
		for (int k = 0; k < sigs; k++) {
			Signal sig;
			Style style;
			XYItemRenderer rend;

			sig = this.getSignal(k);

			// values
			style = sig.getStyle();
			this.plot.setRenderer(k + 1, this.getRenderer(style));
			this.plot.setDataset(k + 1, this.getDataset(sig.getName(), style, sig.getValues()));

			// selected values
			style = sig.getSelectStyle();
			rend = this.getRenderer(style);
			rend.setSeriesVisibleInLegend(0, false);
			this.plot.setRenderer(k + 1 + sigs, rend);
			this.plot.setDataset(k + 1 + sigs,
					this.getDataset(sig.getName() + "_Selected", style, sig.getSelectedValues()));
		}

		// remove old datasets
		for (int k = 1 + (2 * sigs); k < this.plot.getDatasetCount(); k++) {
			this.plot.setRenderer(k, null);
			this.plot.setDataset(k, null);
		}

		this.updateBounds();

		this.plot.getDomainAxis().setLabel(this.getXLabel());
		this.plot.getRangeAxis().setLabel(this.getYLabel());
	}

	/**
	 * Sets the GridLine Visibility
	 * 
	 * @param gridline
	 *            boolean
	 */
	public void setGrid(boolean gridline) {

		this.plot.setDomainGridlinesVisible(gridline);
		// this.plot.setDomainMinorGridlinesVisible(gridline);

		this.plot.setRangeGridlinesVisible(gridline);
		// this.plot.setRangeMinorGridlinesVisible(gridline);

	}

	@Override
	public void save(File f, int width, int height) throws IOException {
		String fileType = "";
		int index;
		AbstractVectorGraphicsIO g2d = null;
		Dimension dim;

		index = f.getName().lastIndexOf('.');
		if (index > 0) {
			fileType = f.getName().substring(index + 1);
		}

		if (fileType.equalsIgnoreCase("eps") || fileType.equalsIgnoreCase("svg") || fileType.equalsIgnoreCase("pdf")) {
			// Color c = (Color) this.chart.getBackgroundPaint();
			//
			// this.chart.setBackgroundPaint(Color.WHITE);

			OutputStream os;
			os = new BufferedOutputStream(new FileOutputStream(f));
			dim = new Dimension(width, height);
			if (fileType.equalsIgnoreCase("eps")) {
				g2d = new PSGraphics2D(os, dim);
			} else if (fileType.equalsIgnoreCase("svg")) {
				g2d = new SVGGraphics2D(os, dim);
			} else if (fileType.equalsIgnoreCase("pdf")) {
				g2d = new PDFGraphics2D(os, dim);
			}
			Rectangle2D rect = new Rectangle2D.Double(0, 0, width, height);
			g2d.startExport();
			this.chart.draw(g2d, rect);
			g2d.endExport();
			os.close();
			// this.chart.setBackgroundPaint(c);
		} else {
			super.save(f, width, height);
		}
	}

	/**
	 * Get 2d-chart of jFreeChart
	 * 
	 * @return 2d-chart
	 */
	public JFreeChart getChart() {
		return this.chart;
	}

	/**
	 * Get ChartPanel of jFreeChart
	 * 
	 * @return 2d-chart
	 */
	public ChartPanel getChartPanel() {
		return this.chartPanel;
	}

	/**
	 * Get plot of jFreeChart
	 * 
	 * @return plot
	 */
	public XYPlot getPlot() {
		return this.plot;
	}

	/**
	 * Get Renderer to style
	 * 
	 * @param style
	 *            Style of signal
	 * @return Renderer for style
	 */
	protected XYItemRenderer getRenderer(Style style) {
		XYItemRenderer renderer = null;

		switch (style.getSignalStyle()) {
		case Line: {
			DefaultXYItemRenderer rend = new DefaultXYItemRenderer();
			rend.setSeriesStroke(0, style.getStroke());
			rend.setSeriesPaint(0, style.getColor());
			rend.setSeriesLinesVisible(0, style.isLineVisible());
			rend.setSeriesShape(0, Marker.getShape(style.getMarker(), style.getMarkerSize()));
			rend.setSeriesShapesVisible(0, style.isMarkerVisible());
			rend.setSeriesShapesFilled(0, style.isMarkerFilled());
			rend.setSeriesVisibleInLegend(0, true);
			rend.setBaseToolTipGenerator(new StandardXYToolTipGenerator());
			renderer = rend;
		}
			break;
		case Bar: {
			XYBarRenderer rend = new XYBarRenderer();
			rend.setSeriesPaint(0, style.getColor());
			rend.setShadowVisible(false);
			renderer = rend;
		}
			break;
		case Step:
			if (style.isMarkerFilled()) {
				renderer = new XYStepAreaRenderer();
			} else {
				renderer = new XYStepRenderer();
			}
			renderer.setSeriesPaint(0, style.getColor());
			break;
		case Block: {
			double min = Double.POSITIVE_INFINITY;
			double max = Double.NEGATIVE_INFINITY;
			double width = Double.POSITIVE_INFINITY;
			double height = Double.POSITIVE_INFINITY;
			XYBlockRenderer rend = new XYBlockRenderer();
			Signal sig;
			double[] x, y;
			double tmp;

			// get min, max
			for (int k = 0; k < this.getNumSignals(); k++) {
				sig = this.getSignal(k);
				if (sig.getStyle().getSignalStyle() == SignalStyle.Block) {
					tmp = sig.getMinZ();
					if (tmp < min) {
						min = tmp;
					}
					tmp = sig.getMaxZ();
					if (tmp > max) {
						max = tmp;
					}
					x = sig.getX();
					y = sig.getY();
					for (int i = 1; i < x.length; i++) {
						tmp = Math.abs(x[i] - x[i - 1]);
						if ((tmp > 0.0) && (tmp < width)) {
							width = tmp;
						}
						tmp = Math.abs(y[i] - y[i - 1]);
						if ((tmp > 0.0) && (tmp < height)) {
							height = tmp;
						}
					}
				}
			}

			// Block size
			rend.setBlockWidth(width);
			rend.setBlockHeight(height);

			ColorPaintScale scale = new ColorPaintScale();
			if (this.zLimitSet) {
				scale.setBound(this.zLimitMin, this.zLimitMax);
			} else {
				scale.setBound(min, max);
			}
			scale.setColorLevels(this.getColorLevels());
			scale.setDirection(this.getColorDirection());
			scale.setTransparency(this.getColorTransparency());
			rend.setPaintScale(scale);

			NumberAxis axis = new NumberAxis();
			if (this.getZLabel() != null) {
				new NumberAxis(this.getZLabel());
			} else {
				new NumberAxis("");
			}

			PaintScaleLegend psc = new PaintScaleLegend(scale, axis);
			psc.setPosition(RectangleEdge.RIGHT);
			this.chart.clearSubtitles();
			this.chart.addSubtitle(psc);

			renderer = rend;
		}
			break;
		case Area: {
			XYAreaRenderer rend = new XYAreaRenderer();
			rend.setSeriesStroke(0, style.getStroke());
			rend.setSeriesPaint(0, style.getColor());
			// rend.setSeriesLinesVisible(0, style.isLineVisible());
			rend.setSeriesShape(0, Marker.getShape(style.getMarker(), style.getMarkerSize()));
			// rend.setSeriesShapesVisible(0, style.isMarkerVisible());
			// rend.setSeriesShapesFilled(0, style.isMarkerFilled());
			rend.setSeriesVisibleInLegend(0, true);
			rend.setBaseToolTipGenerator(new StandardXYToolTipGenerator());
			renderer = rend;
		}
			break;
		case Spline: {
			XYSplineRenderer rend = new XYSplineRenderer();
			rend.setSeriesStroke(0, style.getStroke());
			rend.setSeriesPaint(0, style.getColor());
			rend.setSeriesLinesVisible(0, style.isLineVisible());
			rend.setSeriesShape(0, Marker.getShape(style.getMarker(), style.getMarkerSize()));
			rend.setSeriesShapesVisible(0, style.isMarkerVisible());
			rend.setSeriesShapesFilled(0, style.isMarkerFilled());
			rend.setSeriesVisibleInLegend(0, true);
			rend.setBaseToolTipGenerator(new StandardXYToolTipGenerator());
			renderer = rend;
		}
			break;
		}

		return renderer;
	}

	/**
	 * Get dataset of values
	 * 
	 * @param name
	 *            Name of dataset
	 * @param style
	 *            Style of dataset
	 * @param values
	 *            Values of dataset
	 * @return Dataset of values
	 */
	protected XYDataset getDataset(String name, Style style, double[][] values) {
		DefaultXYZDataset dsxyz;
		XYDataset ds = null;

		switch (style.getSignalStyle()) {
		case Bar:
			dsxyz = new DefaultXYZDataset();
			dsxyz.addSeries(name, values);
			ds = new XYBarDataset(dsxyz, style.getLineWidth());
			break;
		default:
			dsxyz = new DefaultXYZDataset();
			dsxyz.addSeries(name, values);
			ds = dsxyz;
			break;
		}

		return ds;
	}

	/**
	 * Get y-axis show zero
	 * 
	 * @return Show zero
	 */
	public boolean getYAxisShowZero() {
		return this.yAxisShowZero;
	}

	/**
	 * Set y-axis show zero
	 * 
	 * @param showZero
	 *            Show-zero
	 */
	public void setYAxisShowZero(boolean showZero) {
		this.yAxisShowZero = showZero;
		this.zoom();
	}

	@Override
	protected boolean isVectorGraphicPossible() {
		return true;
	}

	/**
	 * Get alpha component of color for transparency
	 * 
	 * @return Alpha
	 */
	public float getColorTransparency() {
		return this.colorAlpha;
	}

	/**
	 * Set alpha component of color for transparency
	 * 
	 * @param alpha
	 *            Alpha
	 */
	public void setColorTransparency(float alpha) {
		this.colorAlpha = alpha;
		this.redraw();
	}

	/**
	 * Add x-axis
	 * 
	 * @param label
	 *            Label of axis
	 */
	public void addXAxis(String label) {
		this.addXAxis(label, 0.0, 1.0);
	}

	/**
	 * Add x-axis
	 * 
	 * @param label
	 *            Label of axis
	 * @param shift
	 *            Shift value (before scale)
	 * @param scale
	 *            Scale factor
	 */
	public void addXAxis(String label, double shift, double scale) {
		ValueAxis axis1, axis2;

		axis1 = this.plot.getRangeAxis(0);
		axis2 = new NumberAxis(label);
		axis2.setLabelFont(axis1.getLabelFont());
		axis2.setTickLabelFont(axis1.getTickLabelFont());
		this.plot.setDomainAxis(this.plot.getDomainAxisCount(), axis2);
		this.domainAxisShiftScale.add(new double[] { shift, scale });
		this.updateBounds();
	}

	/**
	 * Remove x-axis
	 * 
	 * @param index
	 *            Index of axis (0 not allowed)
	 */
	public void rmXAxis(int index) {
		if ((0 < index) && (index < this.domainAxisShiftScale.size())) {
			this.plot.setDomainAxis(index, null);
			this.domainAxisShiftScale.remove(index);
		}
	}

	/**
	 * Set shift and scale-factor of x-axis
	 * 
	 * @param index
	 *            Index of axis
	 * @param shift
	 *            Shift value (before scale)
	 * @param scale
	 *            Scale factor
	 */
	public void setXAxisShiftScale(int index, double shift, double scale) {
		if (index < this.domainAxisShiftScale.size()) {
			this.domainAxisShiftScale.get(index)[0] = shift;
			this.domainAxisShiftScale.get(index)[1] = scale;
		}
		this.updateBounds();
	}

	/**
	 * Get number of x-axis
	 * 
	 * @return Number of axis
	 */
	public int getXAxisCount() {
		return this.plot.getDomainAxisCount();
	}

	/**
	 * Add y-axis
	 * 
	 * @param label
	 *            Label of axis
	 */
	public void addYAxis(String label) {
		this.addYAxis(label, 0.0, 1.0);
	}

	/**
	 * Add y-axis
	 * 
	 * @param label
	 *            Label of axis
	 * @param shift
	 *            Shift value (before scale)
	 * @param scale
	 *            Scale factor
	 */
	public void addYAxis(String label, double shift, double scale) {
		ValueAxis axis1, axis2;

		axis1 = this.plot.getRangeAxis(0);
		axis2 = new NumberAxis(label);
		axis2.setLabelFont(axis1.getLabelFont());
		axis2.setTickLabelFont(axis1.getTickLabelFont());
		this.plot.setRangeAxis(this.plot.getRangeAxisCount(), axis2);
		this.rangeAxisShiftScale.add(new double[] { shift, scale });
		this.updateBounds();
	}

	/**
	 * Remove y-axis
	 * 
	 * @param index
	 *            Index of axis (0 not allowed)
	 */
	public void rmYAxis(int index) {
		if ((0 < index) && (index < this.rangeAxisShiftScale.size())) {
			this.plot.setRangeAxis(index, null);
			this.rangeAxisShiftScale.remove(index);
		}
	}

	/**
	 * Set shift and scale-factor of y-axis
	 * 
	 * @param index
	 *            Index of axis
	 * @param shift
	 *            Shift value (before scale)
	 * @param scale
	 *            Scale factor
	 */
	public void setYAxisShiftScale(int index, double shift, double scale) {
		if (index < this.rangeAxisShiftScale.size()) {
			this.rangeAxisShiftScale.get(index)[0] = shift;
			this.rangeAxisShiftScale.get(index)[1] = scale;
		}
		this.updateBounds();
	}

	/**
	 * Get number of y-axis
	 * 
	 * @return Number of axis
	 */
	public int getYAxisCount() {
		return this.plot.getRangeAxisCount();
	}
}
