import warnings
import numpy as np
from lcm.utils.pytperiodic import PyTPeriodicMatrix


def normMatrix(xc: np.ndarray) -> np.ndarray:
    """
    Calculate a normalized transformation matrix

    Parameters
    ----------
    xc : np.ndarray

    Returns
    -------
    t : np.ndarray
    """
    pp = xc.max(axis=1) - xc.min(axis=1)

    if np.count_nonzero(pp == 0) > 0:
        '''if zero entries in diagonal of pp'''
        warnings.warn('Normalization unsucessful!')
        pp = np.ones_like(pp)

    t = np.diag(1 / pp)

    if np.count_nonzero(t.diagonal() == 0) > 0:
        warnings.warn('Normalization unsucessful!')
        t = np.eye(xc.shape[0])

    return t


def normalization(xc):
    """
    Parameters
    ----------

    xc : (x_dim, n_centers) PyTPeriodicMatrix / np.ndarray
        RBF center coordinates.

    Returns
    -------
    t : (n_centers, n_centers) PyTPeriodicMatrix / np.ndarray
        Transformation matrix

    """
    if isinstance(xc, PyTPeriodicMatrix):
        xc = xc.mean()
        mat = normMatrix(xc)
        t = PyTPeriodicMatrix(mat, mode="dc")
    elif isinstance(xc, np.ndarray):
        t = normMatrix(xc)
    else:
        raise TypeError("Wrong type: only PyTPeriodicMatrix or np.ndarray allowed")

    return t


def qr_preconditioner(xc_per):
    """
    Calculate preconditioning matrix using orthonormalization

    QR-Decomposition calculates Q,R;
    Q is unitary (Q*Q^T = I), Q are orthonormalized vectors
    R is an upper triangular matrix
    A=Q*R
    Q=A*inv(R)
    inv(R) is our transformation matrix
    t=inv(R).transpose() for convinience
    -> Q^T = t * A^T


    Parameters
    ----------
    xc_per : (x_dim, n_centers) PyTPeriodicMatrix
        RBF centers, (corresponds to 'A')
    zc_per :  (z_dim, n_centers) PyTPeriodicMatrix
        RBF center values

    Returns
    -------
    t : (x_dim, x_dim) ndarray
        preconditioning matrix
    """

    xc_mean = xc_per.mean()

    q, r = np.linalg.qr(xc_mean.transpose())
    t = np.linalg.inv(r).transpose()

    return PyTPeriodicMatrix(t, mode="dc")
