# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1024 $
$LastChangedDate:: 2020-06-25 18:45:41 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import unittest
import numpy as np
import numpy.testing as npt
import lcm.math.rbf as rbf
import lcm.math.rbf.kernels as kernels
from lcm.math.rbf import transformations
from lcm.math.rbf.utils.analytic_RBF import AnalyticFunction


class TestRBF(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.plot = False
        xi1 = np.array([np.linspace(0, 10, 5), np.linspace(-10, 10, 5)])
        xi2 = np.array([np.sin(np.linspace(0, 10, 5)), np.cos(np.linspace(-10, 10, 5))])
        md = 0.5
        cls.ref_data = {'xi1-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction(10, 10)),
                        'xi2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction(10, 10)),
                        'xi1-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction(10, 10)),
                        'xi2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction(10, 10))}

    def test__call__(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xc = fnc.getXc()
                zc = fnc.getZc()
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                t = transformations.normalization(xc)
                rbf_interp = rbf.RBF(xc, zc, kernel, max_dist=data['maxdist'], t=t)

                ''' rbf result'''
                zi = rbf_interp(xi)

                ''' analytic result'''
                zi_analytic = fnc.getZi(xi[0, :], xi[1, :])

                '''error'''
                rtol = 2e-1
                atol = 1e-2

                if self.plot:
                    fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(zi - zi_analytic))
                max_error = np.linalg.norm(zi_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    def test_fromMatrices(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xc = fnc.getXc()
                zc = fnc.getZc()
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                t = transformations.normalization(xc)
                tmp_rbf_interp = rbf.RBF(xc, zc, kernel, max_dist=data['maxdist'], t=t)
                rbf_interp = rbf.RBF.fromMatrices(xc, tmp_rbf_interp.c, tmp_rbf_interp.d, kernel, t, max_dist=data['maxdist'])

                ''' rbf result'''
                zi = rbf_interp(xi)

                ''' analytic result'''
                zi_analytic = fnc.getZi(xi[0, :], xi[1, :])

                '''error'''
                rtol = 2e-1
                atol = 1e-2

                if self.plot:
                    fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(zi - zi_analytic))
                max_error = np.linalg.norm(zi_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    @unittest.skip
    def test_xdim(self):
        self.fail()

    @unittest.skip
    def test_zdim(self):
        self.fail()

    def test_eval_dz_dx(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                xc = fnc.getXc()
                zc = fnc.getZc()
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                t = transformations.normalization(xc)
                rbf_interp = rbf.RBF(xc, zc, kernel, max_dist=data['maxdist'], t=t)

                ''' get jacobian matrix ( derivation along x-dimensions)'''
                dzi_dx = rbf_interp.eval_dz_dx(xi)

                ''' analytic result'''
                dzi_dx_analytic = fnc.get_dZi_dx(xi[0, :], xi[1, :])

                '''error'''
                rtol = 1e-1
                atol = 1e-2

                if self.plot:
                    fnc.plotJacobiComparison(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(dzi_dx - dzi_dx_analytic))
                max_error = np.linalg.norm(dzi_dx_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)

    @unittest.skip
    def test_to_msgpack(self):
        self.fail()

    @unittest.skip
    def test_write_msgpack(self):
        self.fail()

    def test_toCBOR(self):
        f = 1
        for key, data in self.ref_data.items():
            with self.subTest(msg='{}-f={}'.format(key, f)):
                fnc = data['fnc']
                xc = fnc.getXc()
                zc = fnc.getZc()
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                rbf_interp = rbf.RBF(xc, zc, kernel, max_dist=data['maxdist'], t=None)

                cbor_obj = rbf_interp.toCBOR()

                self.assertEqual(cbor_obj['type'], '__periodicRBF__')
                self.assertEqual(cbor_obj['max_dist'], rbf_interp.max_dist)
                self.assertEqual(cbor_obj['kernel'], rbf_interp.kernel.id)
                npt.assert_array_equal(cbor_obj['params'], rbf_interp.kernel.params)

                self.assertEqual(cbor_obj['xi_per']['f'], f)
                self.assertEqual(cbor_obj['t_per']['f'], f)
                self.assertEqual(cbor_obj['c_per']['f'], f)
                self.assertEqual(cbor_obj['d_per']['f'], f)
