/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;

import at.lcm.bu21.magopt.structure.Parameter.ParameterPath;

/**
 * Interface of container
 * 
 * @author Werner Koppelsttter
 */
public interface IParamContainer extends IParameter {

	/**
	 * Check if function is enable
	 * 
	 * @return Is function enable
	 * @throws RemoteException
	 */
	public boolean isEnable() throws RemoteException;

	/**
	 * Get if container is enable
	 * 
	 * @return Enable
	 * @throws RemoteException
	 */
	public boolean getEnable() throws RemoteException;

	/**
	 * Set container to given value
	 * 
	 * @param enable
	 *            Enable value
	 * @throws RemoteException
	 */
	public void setEnable(boolean enable) throws RemoteException;

	/**
	 * Get access type of container
	 * 
	 * @return Access state
	 * @throws RemoteException
	 */
	public Access getAccess() throws RemoteException;

	/**
	 * Set access type of container
	 * 
	 * @param access
	 *            Access state
	 * @throws RemoteException
	 */
	public void setAccess(Access access) throws RemoteException;

	/**
	 * Get current incomplete status
	 * 
	 * @return incomplete status
	 * @throws RemoteException
	 */
	public boolean getIncomplete() throws RemoteException;

	/**
	 * Get incomplete status of all datasets
	 * 
	 * @return list of incomplete status
	 * @throws RemoteException
	 */
	public boolean[] getAllIncomplete() throws RemoteException;

	/**
	 * Get List of child-parameter
	 * 
	 * @return parameter-list
	 * @throws RemoteException
	 */
	public IParameter[] getParam() throws RemoteException;

	/**
	 * List all names of public-interfaces
	 * 
	 * @return List of names
	 * @throws RemoteException
	 */
	public String[] listPublicInterfaceName() throws RemoteException;

	/**
	 * Add public-interface if not exists
	 * 
	 * @param name
	 *            Name of interface
	 * @throws RemoteException
	 */
	public void addPublicInterface(String name) throws RemoteException;

	/**
	 * Set public-interface
	 * 
	 * @param name
	 *            Name of interface
	 * @param param
	 *            Parameter-string
	 * @throws RemoteException
	 */
	public void setPublicInterface(String name, String param) throws RemoteException;

	/**
	 * Rename public-interface if exists with dependencies
	 * 
	 * @param oldName
	 *            Old name of interface
	 * @param newName
	 *            New name of interface
	 * @throws RemoteException
	 */
	public void renamePublicInterface(String oldName, String newName) throws RemoteException;

	/**
	 * Remove public-interface
	 * 
	 * @param name
	 *            Name of interface
	 * @return Name of parameter
	 * @throws RemoteException
	 */
	public String removePublicInterface(String name) throws RemoteException;

	/**
	 * Get parameter-name of public-interface
	 * 
	 * @param name
	 *            Name of interface
	 * @return Name of parameter
	 * @throws RemoteException
	 */
	public String getPublicInterfaceParam(String name) throws RemoteException;

	/**
	 * Get path to selected parameter
	 * 
	 * @param iname
	 *            Name of interface
	 * @return Path to parameter
	 * @throws RemoteException
	 */
	// FIXME RMI
	public ParameterPath getPublicInterfacePath(String iname) throws RemoteException;

	/**
	 * Get component property of container
	 * 
	 * @return Component
	 * @throws RemoteException
	 */
	public IComponent getComponent() throws RemoteException;

	/**
	 * Add Component property of container
	 * 
	 * @throws RemoteException
	 */
	public void addComponent() throws RemoteException;

	/**
	 * Remove component property of container
	 * 
	 * @throws RemoteException
	 */
	public void removeComponent() throws RemoteException;

	/**
	 * Convert to class ParamContainer
	 * 
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamContainer convertToParamContainer() throws RemoteException;

	/**
	 * Convert to class ParamMaterial
	 * 
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamMaterial convertToParamMaterial() throws RemoteException;

	/**
	 * Convert to class ParamCollector
	 * 
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamCollector convertToParamCollector() throws RemoteException;

	/**
	 * Convert to class ParamMatlabFunction
	 * 
	 * @param classPath
	 *            Path of matlab class
	 * @param className
	 *            Name of matlab class
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamFunction convertToParamMatlabFunction(String classPath, String className) throws RemoteException;

	/**
	 * Convert to class ParamPythonFunction
	 * 
	 * @param classPath
	 *            Path of python class
	 * @param className
	 *            Name of python class
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamFunction convertToParamPythonFunction(String classPath, String className) throws RemoteException;

	/**
	 * Convert to class ParamJavaFunction
	 * 
	 * @param classPath
	 *            Path of java class
	 * @param className
	 *            Name of java class
	 * @return Converted object
	 * @throws RemoteException
	 */
	public ParamFunction convertToParamJavaFunction(String classPath, String className) throws RemoteException;

	@Override
	public IParamContainer getClientImpl() throws RemoteException;
}
