/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;

import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_29.
 * 
 * - Bug in cloneParam of ParamDataStruct
 * 
 * @author Werner Koppelsttter
 */
public class Converter_29 {

	/** Date-formatter for reading and writing date */
	protected static SimpleDateFormat sdf = new SimpleDateFormat(
			"dd-MM-yyyy HH:mm");

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @return Changed data node
	 */
	static XMLNode convert(XMLNode node) {

		// check parameter identifier
		Converter_29.convertPId(node);

		// check dataset identifier
		Converter_29.convertDsId(node);

		return node;
	}

	/**
	 * Convert parameter identifier
	 * 
	 * @param node
	 *            Data to convert
	 */
	private static void convertPId(XMLNode node) {
		LinkedList<String> list;
		LinkedList<XMLNode> duplicate;
		long next;

		// collect pid
		list = new LinkedList<String>();
		duplicate = new LinkedList<XMLNode>();
		next = 1 + Converter_29.collectPId(node, list, duplicate);

		if (Long.parseLong(node.getNode(2).getNode(1).getText()) > next) {
			next = Long.parseLong(node.getNode(2).getNode(1).getText());
		}

		if (!duplicate.isEmpty()) {
			next = Converter_29.convertDuplicatePId(node, next, duplicate);
		}

		// update next parameter identifier
		node.getNode(2).getNode(1).setText(Long.toString(next));

	}

	/**
	 * Collect parameter-identifier and duplicate identifier
	 * 
	 * @param node
	 *            Node of project
	 * @param list
	 *            List of parameter-identifier
	 * @param duplicate
	 *            List of duplicate identifiers
	 * @return Max used Identifier
	 */
	private static long collectPId(XMLNode node, LinkedList<String> list,
			LinkedList<XMLNode> duplicate) {
		String str;
		long pid = 0, tmp;

		if (node.getName().compareTo("Parameter") == 0) {
			str = node.getNode(0).getText();
			if (list.contains(str)) {
				duplicate.add(node);
			}
			list.add(str);
			pid = Long.parseLong(str, 36);
		}

		// next node
		for (XMLNode n : node.getNodes()) {
			tmp = Converter_29.collectPId(n, list, duplicate);
			if (tmp > pid) {
				pid = tmp;
			}
		}

		return pid;
	}

	/**
	 * Convert duplicate dataset-identifiers
	 * 
	 * @param node
	 *            Node of project
	 * @param next
	 *            Next dataset-identifiers
	 * @param duplicate
	 *            List of duplicate identifiers
	 * @return New next dataset-identifiers
	 */
	private static long convertDuplicatePId(XMLNode node, long next,
			LinkedList<XMLNode> duplicate) {
		XMLNode ps;
		XMLNode[] items;
		String oldPId, newPId;
		LinkedList<String> pids;
		String[] parents;
		String[] par;

		pids = new LinkedList<String>();
		items = node.getNode(2).getNode(4).getNodes();

		// check all duplicate param-ids
		for (XMLNode dup : duplicate) {
			// generate new id
			oldPId = dup.getNode(0).getText();
			pids.add(oldPId);
			newPId = Long.toString(next++, 36);
			pids.add(newPId);
			dup.getNode(0).setText(newPId);

			for (XMLNode ds : items) {
				// set error flag
				ps = Converter_29.setFlag(ds, oldPId, "ERROR", "true");
				// clone paramset
				ps = ps.deepClone();
				ps.setAttribute("pid", newPId);
				ds.getNode(2).addNode(ps);
			}
		}

		// detect parents
		parents = new String[pids.size()];
		for (int k = 0; k < parents.length; k++) {
			parents[k] = Converter_29.getParent(node, pids.get(k), "");
		}

		// add error event for all datasets
		for (XMLNode ds : items) {
			Converter_29.addErrorEvent(ds,
					"Duplicate parameter identifier! Pleast check parameters!",
					node.getNode(0).getNode(0).getNode(1).getText());
			// set incomplete Flags
			for (String p : parents) {
				par = p.split("_");
				for (int k = 1; k < par.length; k++) {
					Converter_29.setFlag(ds, par[k], "INCOMPLETE", "true");
				}
			}
		}

		return next;
	}

	/**
	 * Collect parameter-identifier and duplicate identifier
	 * 
	 * @param node
	 *            Node of project
	 * @param pid
	 *            Parameter-identifier to search
	 * @param parents
	 *            List of parents
	 * @return Parents of node of pid
	 */
	private static String getParent(XMLNode node, String pid, String parents) {
		String par = null;

		if (node.getName().compareTo("Parameter") == 0) {
			if (pid.compareTo(node.getNode(0).getText()) == 0) {
				par = parents;
			}
		} else {
			// call of super class
			par = Converter_29.getParent(node.getNode(0), pid, parents);

			if (node.getName().compareTo("ParamContainer") == 0) {
				// add parent
				parents += "_" + node.getNode(0).getNode(0).getText();
				// next param
				for (XMLNode n : node.getNode(1).getNodes()) {
					if (par == null) {
						par = Converter_29.getParent(n, pid, parents);
					}
				}
			}
		}

		return par;
	}

	/**
	 * Convert dataset identifier
	 * 
	 * @param node
	 *            Data to convert
	 */
	private static void convertDsId(XMLNode node) {
		LinkedList<String> list, duplicate;
		long next;

		// paramter-data-struct exists
		if (node.getNodeSize() >= 2) {
			// collect dsid
			list = new LinkedList<String>();
			duplicate = new LinkedList<String>();
			next = 1 + Converter_29.collectDsId(node, list, duplicate);
			if (Long.parseLong(node.getNode(2).getNode(2).getText()) > next) {
				next = Long.parseLong(node.getNode(2).getNode(2).getText());
			}

			if (!duplicate.isEmpty()) {
				next = Converter_29.convertDuplicateDsId(node, next, duplicate);
			}

			// update next dataset identifier
			node.getNode(2).getNode(2).setText(Long.toString(next));
		}
	}

	/**
	 * Collect dataset-identifier and duplicate identifier
	 * 
	 * @param node
	 *            Node of project
	 * @param list
	 *            List of dataset-identifier
	 * @param duplicate
	 *            List of duplicate identifiers
	 * @return Max used Identifier
	 */
	private static long collectDsId(XMLNode node, LinkedList<String> list,
			LinkedList<String> duplicate) {
		String str;
		long dsid = 0, tmp;
		XMLNode items;

		items = node.getNode(2).getNode(3);
		for (XMLNode n : items.getNodes()) {
			str = n.getText();
			if (list.contains(str)) {
				duplicate.add(str);
			}
			list.add(str);
			tmp = Long.parseLong(str, 36);
			if (tmp > dsid) {
				dsid = tmp;
			}
		}
		return dsid;
	}

	/**
	 * Convert duplicate dataset-identifiers
	 * 
	 * @param node
	 *            Node of project
	 * @param next
	 *            Next dataset-identifiers
	 * @param duplicate
	 *            List of duplicate identifiers
	 * @return New next dataset-identifiers
	 */
	private static long convertDuplicateDsId(XMLNode node, long next,
			LinkedList<String> duplicate) {
		XMLNode items, ds;
		boolean first;
		String dsid;

		// check all duplicate ds-ids
		items = node.getNode(2).getNode(3);
		for (String key : duplicate) {
			first = true;
			for (int k = 0; k < items.getNodeSize(); k++) {
				if (items.getNode(k).getText().compareTo(key) == 0) {
					ds = node.getNode(2).getNode(4).getNode(k);
					if (!first) {
						// change datatset identifier
						dsid = Long.toString(next++, 36);
						items.getNode(k).setText(dsid);
						ds.getNode(0).setText(dsid);
					}
					first = false;

					// set error flag
					Converter_29.setFlag(ds, node.getNode(0).getNode(0)
							.getNode(0).getText(), "ERROR", "true");

					// add error event
					Converter_29
							.addErrorEvent(
									ds,
									"Duplicate dataset identifier! Pleast check datasets!",
									node.getNode(0).getNode(0).getNode(1)
											.getText());
				}
			}
		}

		return next;
	}

	/**
	 * Set error-flag of specified parameter-set
	 * 
	 * @param ds
	 *            Dataset of parameter-set
	 * @param pid
	 *            Parameter-identifier
	 * @param flag
	 *            Name of flag
	 * @param value
	 *            Value of flag
	 * @return Modified parameter-set
	 */
	private static XMLNode setFlag(XMLNode ds, String pid, String flag,
			String value) {
		XMLNode ps, error;

		// find paramset for error flag
		ps = null;
		for (XMLNode p : ds.getNode(2).getNodes()) {
			if (pid.compareTo(p.getAttribute("pid")) == 0) {
				ps = p;
			}
		}
		// create paramset if not exists
		if (ps == null) {
			ps = new XMLNode("ParamSet");
			ps.setAttribute("pid", pid);
			ds.getNode(2).addNode(ps);

			ps.addNode(new XMLNode("Data"));
			ps.addNode(new XMLNode("DocFiles"));
		}

		// find error flag
		error = null;
		for (XMLNode data : ps.getNode(0).getNodes()) {
			if (data.getName().compareTo(flag) == 0) {
				error = data;
			}
		}
		// create error if not exists
		if (error == null) {
			error = new XMLNode(flag);
			error.addNode(new XMLNode("Boolean"));
			ps.getNode(0).addNode(error);
		}

		// set error
		error.getNode(0).setText(value);

		return ps;
	}

	/**
	 * Add error event to event-log
	 * 
	 * @param ds
	 *            Dataset of event-log
	 * @param msg
	 *            Event-massage
	 * @param ident
	 *            Source identifier for event
	 */
	private static void addErrorEvent(XMLNode ds, String msg, String ident) {
		XMLNode event, tmp;

		event = new XMLNode("Event");
		ds.getNode(3).getNode(1).addNode(event);

		tmp = new XMLNode("Type");
		tmp.setText("ERROR");
		event.addNode(tmp);

		tmp = new XMLNode("Message");
		tmp.setText(msg);
		event.addNode(tmp);

		tmp = new XMLNode("Date");
		tmp.setText(Converter_29.sdf.format(new Date()));
		event.addNode(tmp);

		tmp = new XMLNode("Identifier");
		tmp.setText(ident);
		event.addNode(tmp);

		tmp = new XMLNode("Stack");
		event.addNode(tmp);
	}

}
