/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Locale;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_32.
 * 
 * - Round variables <br>
 * 
 * @author Werner Koppelsttter
 */
public class Converter_32 implements MopAccess {

	/** Mop-file of project */
	MopFile mop;

	/** Node of load archive */
	XMLNode node = null;

	/**
	 * Convert the given node
	 * 
	 * @param mop
	 *            Mop-file with optimization
	 * @throws Exception
	 */
	static void convertOptimizer(MopFile mop) throws Exception {
		Converter_32 convert;
		String[] filelist;

		convert = new Converter_32(mop);
		filelist = mop.getFileNames();
		for (String name : filelist) {
			if (name.startsWith("GeneticOptimization/")
					&& name.endsWith(".xml")) {
				convert.convertOpt(name);
			}
		}
	}

	/**
	 * Convert number to string
	 * 
	 * @param val
	 *            Value to convert
	 * @param precision
	 *            Precision of value
	 * @return String representation of value
	 */
	static String num2str(double val, int precision) {
		if (precision > 0) {
			return Double
					.valueOf(
							String.format(Locale.ENGLISH, "%."
									+ (precision - 1) + "e", val)).toString()
					.replace('E', 'e');
		} else {
			return Double.toString(val);
		}
	}

	/**
	 * Create converter object
	 * 
	 * @param mop
	 *            Mop-file to convert
	 */
	private Converter_32(MopFile mop) {
		this.mop = mop;
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	private void convertOpt(String name) throws Exception {
		this.node = null;
		this.mop.readStream(this, name);
		this.convert(this.node);
		this.mop.writeStream(this, name);
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param node
	 *            Archive to convert
	 */
	private void convert(XMLNode node) {
		XMLNode individuals;

		for (XMLNode n : node.getNodes()) {
			if (n.getAttribute("name").compareTo("memory") == 0) {
				individuals = n.getNode(0).getNode(0);
				for (XMLNode ind : individuals.getNodes()) {
					this.convertIndividual(ind);
				}
			}
		}

	}

	/**
	 * Convert individual
	 * 
	 * @param node
	 *            individual node
	 */
	private void convertIndividual(XMLNode node) {
		String str, oldStr, newStr, s;
		double val;
		int index;

		// find paretos
		for (XMLNode n : node.getNodes()) {
			if (n.getAttribute("name").compareTo("var") == 0) {
				oldStr = n.getNode(0).getText();

				if (oldStr.startsWith("[")) {
					newStr = "[";
					oldStr = oldStr.substring(1, oldStr.length() - 1);
				} else {
					newStr = "";
				}

				while (!oldStr.isEmpty()) {
					index = oldStr.indexOf(" ");
					if (index >= 0) {
						str = oldStr.substring(0, index);
						oldStr = oldStr.substring(index + 1);
					} else {
						str = oldStr;
						oldStr = "";
					}

					val = Double.parseDouble(str);
					s = Converter_32.num2str(val, 14);
					if ((str.compareTo(s) != 0)
							&& (s.compareTo(Converter_32.num2str(val, 10)) == 0)) {
						str = s;
					}
					newStr += str + " ";
				}
				newStr = newStr.substring(0, newStr.length() - 1);

				if (newStr.startsWith("[")) {
					newStr += "]";
				}

				n.getNode(0).setText(newStr);
			}
		}
	}

	@Override
	public void readStream(InputStream is) throws Exception {
		XMLStreamReader xmlReader = XMLInputFactory.newInstance()
				.createXMLStreamReader(is, "utf-8");
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		// read objects
		this.node = new XMLNode(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new XMLStreamException("Undefined error by reading file!");
		}
		xmlReader.close();
	}

	@Override
	public void writeStream(OutputStream os) throws Exception {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance()
				.createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");

		// write object to File
		this.node.writeXML(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}
}
