/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.math.BigInteger;

import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_8.
 * 
 * - Rename Comment to IndexComment <br>
 * - Rename ParamFieldFnc to ParamFormula <br>
 * - Remove ReadOnly and Used flags <br>
 * - Remove MD5-String of ParamFnc <br>
 * - Change checksum format and string <br>
 * - Update ParamFnc <br>
 * - Update object <br>
 * - Add Comment
 * 
 * @author Werner Koppelsttter
 */
class Converter_8 {

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @return Changed data node
	 */
	static XMLNode convert(XMLNode node) {

		// Convert Parameter
		Converter_8.renameNodes(node);
		Converter_8.removeNodes(node);
		Converter_8.updateCheckSum(node);
		Converter_8.updateParamFnc(node);
		Converter_8.updateObject(node);
		Converter_8.addComment(node);

		return node;
	}

	/**
	 * Recursive rename of ParamFieldFnc and Comment
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void renameNodes(XMLNode node) {
		String attr;

		if (node.getName().compareTo("Comment") == 0) {
			node.setName("DSComment");
		}

		if (node.getName().compareTo("ParamFieldFnc") == 0) {
			node.setName("ParamFormula");
		}

		attr = node.getAttribute("object");
		if ((attr != null) && (attr.compareTo("ParamFieldFnc") == 0)) {
			node.setAttribute("object", "ParamFormula");
		}

		// next Node
		for (XMLNode n : node.getNodes()) {
			Converter_8.renameNodes(n);
		}
	}

	/**
	 * Recursive update of node to remove flags
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void removeNodes(XMLNode node) {
		XMLNode item;

		// next Node
		int i = 0;
		while (i < node.getNodeSize()) {
			item = node.getNode(i);
			if (item.getName().compareTo("ReadOnly") == 0) {
				// Remove ReadOnly-flag
				node.rmNode(i);
			} else if (item.getName().compareTo("Used") == 0) {
				// Remove Used-flag
				node.rmNode(i);
			} else if ((item.getName().compareTo("MD5") == 0)
					&& (node.getName().compareTo("ParamFnc") == 0)) {
				// Remove MD5-String from ParamFnc
				node.rmNode(i);
			} else {
				// recursive call
				Converter_8.removeNodes(item);
				i++;
			}
		}
	}

	/**
	 * Update Checksum format
	 * 
	 * @param node
	 */
	private static void updateCheckSum(XMLNode node) {
		XMLNode item, tmp;

		// next Node
		int i = 0;
		while (i < node.getNodeSize()) {
			item = node.getNode(i);
			if (item.getName().compareTo("MD5") == 0) {
				item.setName("CheckSum");
				for (int j = 0; j < item.getNodeSize(); j++) {
					tmp = item.getNode(j);
					if (tmp.getText() != null) {
						try {
							tmp.setText((new BigInteger(tmp.getText()))
									.toString(16));
						} catch (Exception ex) {
							tmp.setText("");
						}
					}
				}
			} else {
				// recursive call
				Converter_8.updateCheckSum(item);
				i++;
			}
		}
	}

	/**
	 * Recursive update of ParamFnc
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void updateParamFnc(XMLNode node) {
		XMLNode item, tmp, className, data;

		// next Node
		int i = 0;
		while (i < node.getNodeSize()) {
			item = node.getNode(i);
			if (item.getName().compareTo("ParamGeometry") == 0) {
				// ParamMatlabFunction
				item.setName("ParamMatlabFunction");
				className = new XMLNode("ClassName");
				className.setText(item.getAttribute("object"));
				item.addNode(className);
				// data in ParamFnc
				data = item.rmNode(1);
				item = item.getNode(0);
				item.addNode(2, data);
				item.setName("ParamFunction");
			} else if (item.getName().compareTo("ParamFnc") == 0) {
				item.setName("ParamFunction");
				// ParamMatlabFunction
				tmp = new XMLNode("ParamMatlabFunction");
				node.setNode(i, tmp);
				tmp.addNode(item);
				tmp.setAttribute("name", item.getAttribute("name"));
				tmp.setAttribute("object", item.getAttribute("object"));
				className = new XMLNode("ClassName");
				className.setText(item.getAttribute("object"));
				tmp.addNode(className);
				// data in ParamFnc
				data = new XMLNode("Data");
				for (int j = 0; j < item.getNode(2).getNodeSize(); j++) {
					tmp = new XMLNode("Numeric");
					tmp.setText("[]");
					data.addNode(tmp);
				}
				item.addNode(2, data);
			}

			// recursive call
			Converter_8.updateParamFnc(item);
			i++;
		}

	}

	/**
	 * Recursive update of object attribute
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void updateObject(XMLNode node) {

		String attr = node.getAttribute("object");
		if (attr != null) {
			if ((attr.compareTo("Parameter") != 0)
					&& (attr.compareTo("ParamField") != 0)
					&& (attr.compareTo("ParamFormula") != 0)
					&& (attr.compareTo("ParamContainer") != 0)
					&& (attr.compareTo("Project") != 0)
					&& (attr.compareTo("ParamDirectory") != 0)
					&& (attr.compareTo("ParamFile") != 0)
					&& (attr.compareTo("ParamMaterial") != 0)) {
				node.setAttribute("object", "ParamMatlabFunction");
			}
		}

		// next Node
		for (XMLNode n : node.getNodes()) {
			// recursive call
			Converter_8.updateObject(n);
		}

	}

	/**
	 * Recursive add comment to Parameter
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void addComment(XMLNode node) {

		if (node.getName().compareTo("Parameter") == 0) {
			node.addNode(2, new XMLNode("Comment"));
		}

		// next Node
		for (XMLNode n : node.getNodes()) {
			// recursive call
			Converter_8.addComment(n);
		}

	}
}
