/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.event;

import java.util.EventObject;

/**
 * Defines an event that encapsulates changes of event-log
 *
 * @author Werner Koppelsttter
 */
public class EventLogEvent extends EventObject {

	/**
	 * Identifies the type of event
	 *
	 * @author Werner Koppelsttter
	 */
	public enum EventType {
		/** Identifies the changes of one or more event entries */
		EVENT_CHANGED,
		/** Identifies the addition of one or more event entries */
		EVENT_ADDED,
		/** Identifies the removal of one or more event entries */
		EVENT_REMOVED
	}

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Type of event */
	private EventType type;

	/** Events of range */
	private Event[] events;

	/** Lover index of range */
	private int from;

	/** Upper index of range */
	private int to;

	/**
	 * Create new event of eventlog change
	 * 
	 * @param source
	 *            Source object <code>EventLog</code>
	 * @param type
	 *            Type of changes
	 * @param events
	 *            Events of range
	 * @param from
	 *            Lower index of range
	 * @param to
	 *            Upper index of range
	 */
	public EventLogEvent(Object source, EventType type, Event[] events,
			int from, int to) {
		super(source);
		this.type = type;
		this.events = events;
		this.from = from;
		this.to = to;
	}

	/**
	 * Get type of event
	 * 
	 * @return Event type
	 */
	public EventType getType() {
		return this.type;
	}

	/**
	 * Get event of range
	 * 
	 * @return Events of range
	 */
	public Event[] getEvents() {
		return this.events;
	}

	/**
	 * Get lower index of interval
	 * 
	 * @return Lower index
	 */
	public int getFromIndex() {
		return this.from;
	}

	/**
	 * Get upper index of interval
	 * 
	 * @return Upper index
	 */
	public int getToIndex() {
		return this.to;
	}

}
