/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;

/**
 * Scan formula and return Tokens (lexical analyze)
 *
 * @author Werner Koppelsttter
 */
public class Scanner {

	/** End of file character */
	static final char eofCh = (char) -1;

	/** Input as reader */
	private Reader in;
	/** Next character */
	private char ch;
	/** Row of next character */
	int line;
	/** Column of next character */
	int col;

	/**
	 * Initialize scanner with formula-string
	 * 
	 * @param formula
	 *            Expression to scan
	 */
	public Scanner(String formula) {
		this(new StringReader(formula));
	}

	/**
	 * Initialize scanner with reader
	 * 
	 * @param reader
	 *            Reader with formula
	 */
	public Scanner(Reader reader) {
		this.in = reader;
		this.line = 1;
		this.col = 0;
		this.nextCh();
	}

	/**
	 * Read next character
	 */
	private void nextCh() {
		try {
			this.ch = (char) this.in.read();
			this.col++;
			if (this.ch == '\n') {
				this.line++;
				this.col = 0;
			}
		} catch (IOException ex) {
			this.ch = Scanner.eofCh;
		}
	}

	/**
	 * Read next token
	 * 
	 * @return Scanned token
	 * @throws ScannerException
	 */
	public Token next() throws ScannerException {
		Token tok;
		while (this.ch <= ' ') {
			this.nextCh();
		}

		tok = new Token();
		tok.line = this.line;
		tok.col = this.col;

		switch (this.ch) {
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'g':
		case 'h':
		case 'i':
		case 'j':
		case 'k':
		case 'l':
		case 'm':
		case 'n':
		case 'o':
		case 'p':
		case 'q':
		case 'r':
		case 's':
		case 't':
		case 'u':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z':
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'G':
		case 'H':
		case 'I':
		case 'J':
		case 'K':
		case 'L':
		case 'M':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'S':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
			this.readName(tok);
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			this.readNumber(tok);
			break;
		case '\"':
			this.readString(tok);
			break;
		case '+':
			this.nextCh();
			tok.kind = TokenKind.PLUS;
			break;
		case '-':
			this.nextCh();
			tok.kind = TokenKind.MINUS;
			break;
		case '*':
			this.nextCh();
			tok.kind = TokenKind.TIMES;
			break;
		case '/':
			this.nextCh();
			tok.kind = TokenKind.SLASH;
			break;
		case '%':
			this.nextCh();
			tok.kind = TokenKind.PERCENT;
			break;
		case '^':
			this.nextCh();
			tok.kind = TokenKind.CARET;
			break;
		case '\'':
			this.nextCh();
			tok.kind = TokenKind.ACCENT;
			break;
		case '=':
			this.nextCh();
			if (this.ch == '=') {
				this.nextCh();
				tok.kind = TokenKind.EQL;
			} else {
				tok.str = "=" + this.ch;
				tok.kind = TokenKind.NONE;
				throw new ScannerException(this, "Single '=' not allowed");
			}
			break;
		case '!':
			this.nextCh();
			if (this.ch == '=') {
				this.nextCh();
				tok.kind = TokenKind.NEQ;
			} else {
				tok.kind = TokenKind.CALL;
			}
			break;
		case '<':
			this.nextCh();
			if (this.ch == '=') {
				this.nextCh();
				tok.kind = TokenKind.LEQ;
			} else {
				tok.kind = TokenKind.LSS;
			}
			break;
		case '>':
			this.nextCh();
			if (this.ch == '=') {
				this.nextCh();
				tok.kind = TokenKind.GEQ;
			} else {
				tok.kind = TokenKind.GTR;
			}
			break;
		case '&':
			this.nextCh();
			if (this.ch == '&') {
				this.nextCh();
				tok.kind = TokenKind.AND;
			} else {
				tok.str = "&" + this.ch;
				tok.kind = TokenKind.NONE;
				throw new ScannerException(this, "Single '&' not allowed");
			}
			break;
		case '|':
			this.nextCh();
			if (this.ch == '|') {
				this.nextCh();
				tok.kind = TokenKind.OR;
			} else {
				tok.str = "|" + this.ch;
				tok.kind = TokenKind.NONE;
				throw new ScannerException(this, "Single '|' not allowed");
			}
			break;
		case ';':
			this.nextCh();
			tok.kind = TokenKind.SEMICOLON;
			break;
		case ',':
			this.nextCh();
			tok.kind = TokenKind.COMMA;
			break;
		case '.':
			this.nextCh();
			if (this.ch == '*') {
				this.nextCh();
				tok.kind = TokenKind.DOTTIMES;
			} else if (this.ch == '/') {
				this.nextCh();
				tok.kind = TokenKind.DOTSLASH;
			} else if (this.ch == '%') {
				this.nextCh();
				tok.kind = TokenKind.DOTPERCENT;
			} else if (this.ch == '^') {
				this.nextCh();
				tok.kind = TokenKind.DOTCARET;
			} else {
				tok.kind = TokenKind.PERIOD;
			}
			break;
		case ':':
			this.nextCh();
			tok.kind = TokenKind.COLON;
			break;
		case '(':
			this.nextCh();
			tok.kind = TokenKind.LPAR;
			break;
		case ')':
			this.nextCh();
			tok.kind = TokenKind.RPAR;
			break;
		case '[':
			this.nextCh();
			tok.kind = TokenKind.LBRACK;
			break;
		case ']':
			this.nextCh();
			tok.kind = TokenKind.RBRACK;
			break;
		case '{':
			this.nextCh();
			tok.kind = TokenKind.LBRACE;
			break;
		case '}':
			this.nextCh();
			tok.kind = TokenKind.RBRACE;
			break;
		case eofCh:
			tok.kind = TokenKind.EOF;
			break;
		default:
			tok.str = "" + this.ch;
			this.nextCh();
			tok.kind = TokenKind.NONE;
			throw new ScannerException(this, "Unexpected sign " + this.ch);
		}

		return tok;
	}

	/**
	 * Read names and keywords
	 * 
	 * @param tok
	 *            Token to insert name
	 */
	private void readName(Token tok) {
		StringBuilder buf = new StringBuilder();
		TokenKind kw;

		buf.append(this.ch);
		this.nextCh();
		while ((('a' <= this.ch) && (this.ch <= 'z')) || (('A' <= this.ch) && (this.ch <= 'Z'))
				|| (('0' <= this.ch) && (this.ch <= '9')) || (this.ch == '_')) {
			buf.append(this.ch);
			this.nextCh();
		}

		tok.str = buf.toString();
		kw = TokenKind.getKeywords().get(tok.str);
		if (kw != null) {
			tok.kind = kw;
			tok = null;
		} else {
			tok.kind = TokenKind.IDENT;
		}
	}

	/**
	 * Read numbers (real or imaginary)
	 * 
	 * @param tok
	 *            Token to insert number
	 * @throws ScannerException
	 */
	private void readNumber(Token tok) throws ScannerException {
		StringBuilder buf = new StringBuilder();
		double d;

		buf.append(this.ch);
		this.nextCh();
		while (('0' <= this.ch) && (this.ch <= '9')) {
			buf.append(this.ch);
			this.nextCh();
		}

		if (this.ch == '.') {
			buf.append(this.ch);
			this.nextCh();
			while (('0' <= this.ch) && (this.ch <= '9')) {
				buf.append(this.ch);
				this.nextCh();
			}
		}

		if ((this.ch == 'e') || (this.ch == 'E')) {
			buf.append(this.ch);
			this.nextCh();
			if ((this.ch == '+') | (this.ch == '-')) {
				buf.append(this.ch);
				this.nextCh();
			}
			do {
				buf.append(this.ch);
				this.nextCh();
			} while (('0' <= this.ch) && (this.ch <= '9'));
		}

		try {
			d = Double.parseDouble(buf.toString());
			if (this.ch == 'i') {
				this.nextCh();
				tok.val = new TNumeric(0.0, d);
			} else {
				tok.val = new TNumeric(d);
			}
			tok.kind = TokenKind.NUMBER;
		} catch (TDataException e) {
			if (this.ch == 'i') {
				buf.append(this.ch);
				this.nextCh();
			}
			tok.str = buf.toString();
			tok.kind = TokenKind.NONE;
			throw new ScannerException(this, "Unable to parse number '" + tok.str + "'");
		}
	}

	/**
	 * Read string (bewtween ' ')
	 * 
	 * @param tok
	 *            Token to insert string
	 * @throws ScannerException
	 */
	private void readString(Token tok) throws ScannerException {
		StringBuilder buf = new StringBuilder();

		this.nextCh();
		while ((this.ch != Scanner.eofCh) && (this.ch != '\"')) {
			buf.append(this.ch);
			this.nextCh();
		}
		tok.str = buf.toString();
		tok.kind = TokenKind.STRING;

		if (this.ch == '\"') {
			this.nextCh();
		} else {
			throw new ScannerException(this, "Missing end of string '" + tok.str + "'");
		}

	}
}
