/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.decmo;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;

import javax.swing.GroupLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.SwingConstants;

import at.lcm.bu21.optim.algorithm.JPanelSolver;

/**
 * Panel for settings of generation-based spea2real optimizer
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 * @author Ciprian Zavoianu+
 */
public class GenDecmoPanel extends JPanelSolver {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Spea2Real optimizer with setting values */
	private GenDecmo solver;

	/** Panel of ssa-solver */
	private JPanelSolver ssaPanel;

	/** Population size */
	private JTextField txtPopulationSize;

	/** Generation timeout */
	private JTextField txtGenerationTimeout;

	/** Panel for steaty-state-asynchron algotithm */
	private JPanel panelSsa;

	/**
	 * Create panel. Used by WindowsBuilder
	 */
	private GenDecmoPanel() {
		this.initComponents();
	}

	/**
	 * Creates new setting-panel for Spea2Real
	 * 
	 * @param solver
	 *            Optimizer with setting values
	 */
	public GenDecmoPanel(GenDecmo solver) {
		this();
		this.solver = solver;

		this.panelSsa.setLayout(new BorderLayout());
		this.ssaPanel = new SsaDecmoPanel(solver);
		this.panelSsa.add(this.ssaPanel);

		this.txtPopulationSize.setText(Integer.toString(this.solver
				.getPopulationSize()));
		this.txtGenerationTimeout.setText(Integer.toString(this.solver
				.getGenerationTimeout()));

		this.refreshEnabled();
	}

	@Override
	public void setSettingEnabled(boolean enable) {
		super.setSettingEnabled(enable);
		this.ssaPanel.setSettingEnabled(enable);
	}

	@Override
	public void refreshEnabled() {
		boolean en;

		en = this.enable && !this.solver.getProblem().isLocked();

		this.txtPopulationSize.setEnabled(en);
		this.txtGenerationTimeout.setEnabled(en);

		this.ssaPanel.refreshEnabled();
	}

	/**
	 * Set value of population-size
	 */
	private void setPopulationSize() {
		try {
			this.solver.setPopulationSize(Integer
					.parseInt(this.txtPopulationSize.getText()));
		} catch (Exception e) {
		}
		this.txtPopulationSize.setText(Integer.toString(this.solver
				.getPopulationSize()));
	}

	/**
	 * Set value of generation-timeout
	 */
	private void setGenerationTimeout() {
		try {
			this.solver.setGenerationTimeout(Integer
					.parseInt(this.txtGenerationTimeout.getText()));
		} catch (Exception e) {
		}
		this.txtGenerationTimeout.setText(Integer.toString(this.solver
				.getGenerationTimeout()));
	}

	/**
	 * Initialize components
	 */
	private void initComponents() {
		JLabel labGenerationTimeoutLabel, labGenerationTimeoutUnit;
		JLabel labPopulationSizeLabel, labPopulationSizeUnit;

		labPopulationSizeLabel = new JLabel();
		this.txtPopulationSize = new JTextField();
		labPopulationSizeUnit = new JLabel();
		labGenerationTimeoutLabel = new JLabel();
		this.txtGenerationTimeout = new JTextField();
		labGenerationTimeoutUnit = new JLabel();
		this.panelSsa = new JPanel();

		labPopulationSizeLabel.setText("Population size:");

		this.txtPopulationSize.setPreferredSize(new Dimension(60, 20));
		this.txtPopulationSize.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				GenDecmoPanel.this.setPopulationSize();
			}
		});
		this.txtPopulationSize.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				GenDecmoPanel.this.setPopulationSize();
			}
		});

		labPopulationSizeUnit.setText("Individuals");

		labGenerationTimeoutLabel.setText("Generation Timeout:");

		this.txtGenerationTimeout.setPreferredSize(new Dimension(60, 20));
		this.txtGenerationTimeout.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				GenDecmoPanel.this.setGenerationTimeout();
			}
		});
		this.txtGenerationTimeout.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				GenDecmoPanel.this.setGenerationTimeout();
			}
		});

		labGenerationTimeoutUnit.setText("min.");

		GroupLayout panelSsaLayout = new GroupLayout(this.panelSsa);
		this.panelSsa.setLayout(panelSsaLayout);
		panelSsaLayout.setHorizontalGroup(panelSsaLayout.createParallelGroup(
				GroupLayout.Alignment.LEADING).addGap(0, 538, Short.MAX_VALUE));
		panelSsaLayout.setVerticalGroup(panelSsaLayout.createParallelGroup(
				GroupLayout.Alignment.LEADING).addGap(0, 223, Short.MAX_VALUE));

		GroupLayout layout = new GroupLayout(this);
		this.setLayout(layout);
		layout.setHorizontalGroup(layout
				.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.LEADING)
												.addComponent(
														labPopulationSizeLabel)
												.addComponent(
														labGenerationTimeoutLabel))
								.addGap(64, 64, 64)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.LEADING,
												false)
												.addComponent(
														this.txtGenerationTimeout,
														0, 60, Short.MAX_VALUE)
												.addComponent(
														this.txtPopulationSize,
														GroupLayout.DEFAULT_SIZE,
														60, Short.MAX_VALUE))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.LEADING)
												.addComponent(
														labGenerationTimeoutUnit)
												.addComponent(
														labPopulationSizeUnit))
								.addContainerGap(245, Short.MAX_VALUE))
				.addComponent(this.panelSsa, GroupLayout.DEFAULT_SIZE,
						GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE));

		layout.linkSize(SwingConstants.HORIZONTAL, new Component[] {
				this.txtGenerationTimeout, this.txtPopulationSize });

		layout.setVerticalGroup(layout
				.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labPopulationSizeLabel)
												.addComponent(
														this.txtPopulationSize,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(
														labPopulationSizeUnit))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labGenerationTimeoutLabel)
												.addComponent(
														this.txtGenerationTimeout,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(
														labGenerationTimeoutUnit))
								.addGap(18, 18, 18)
								.addComponent(this.panelSsa,
										GroupLayout.DEFAULT_SIZE,
										GroupLayout.DEFAULT_SIZE,
										Short.MAX_VALUE)));
	}
}
