/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.decmo;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;

import javax.swing.GroupLayout;
import javax.swing.GroupLayout.Alignment;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.SwingConstants;

import at.lcm.bu21.optim.algorithm.JPanelSolver;

/**
 * Panel for settings of spea2real optimizer
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 * @author Ciprian Zavoianu
 */
public class SsaDecmoPanel extends JPanelSolver {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Spea2Real optimizer with setting values */
	private Decmo solver;

	/** Size of archive */
	private JTextField txtArchiveSize;

	/** Crossover distribution index */
	private JTextField txtCrossoverDistributionIndex;

	/** Crossover probability */
	private JTextField txtCrossoverProbability;

	/** Mutation distribution index */
	private JTextField txtMutationDistributionIndex;

	/** Mutation probability */
	private JTextField txtMutationProbability;

	/** DE crossover-rate */
	private JTextField txtDeCR;

	/** DE mutation factor */
	private JTextField txtDeF;

	/**
	 * Create panel. Used by WindowsBuilder
	 */
	private SsaDecmoPanel() {
		this.initComponents();
	}

	/**
	 * Creates new setting-panel for Spea2Real
	 * 
	 * @param solver
	 *            Optimizer with setting values
	 */
	public SsaDecmoPanel(Decmo solver) {
		this();
		this.solver = solver;

		this.txtArchiveSize.setText(Integer.toString(this.solver
				.getArchiveSize()));
		this.txtCrossoverProbability.setText(Double.toString(this.solver
				.getCrossoverProbability()));
		this.txtCrossoverDistributionIndex.setText(Double.toString(this.solver
				.getCrossoverDistributionIndex()));
		this.txtMutationProbability.setText(Double.toString(this.solver
				.getMutationProbability()));
		this.txtMutationDistributionIndex.setText(Double.toString(this.solver
				.getMutationDistributionIndex()));
		this.txtDeCR.setText(Double.toString(this.solver.getDeCR()));
		this.txtDeF.setText(Double.toString(this.solver.getDeF()));

		this.refreshEnabled();
	}

	@Override
	public void refreshEnabled() {
		boolean en;

		en = this.enable && !this.solver.getProblem().isLocked();

		this.txtArchiveSize.setEnabled(en);
		this.txtCrossoverProbability.setEnabled(en);
		this.txtCrossoverDistributionIndex.setEnabled(en);
		this.txtMutationProbability.setEnabled(en);
		this.txtMutationDistributionIndex.setEnabled(en);
		this.txtDeCR.setEnabled(en);
		this.txtDeF.setEnabled(en);
	}

	/**
	 * Set value of archive-size
	 */
	private void setArchiveSize() {
		try {
			this.solver.setArchiveSize(Integer.parseInt(this.txtArchiveSize
					.getText()));
		} catch (Exception e) {
		}
		this.txtArchiveSize.setText(Integer.toString(this.solver
				.getArchiveSize()));
	}

	/**
	 * Set value of crossover-probability
	 */
	private void setCrossoverProbability() {
		try {
			this.solver.setCrossoverProbability(Double
					.parseDouble(this.txtCrossoverProbability.getText()));
		} catch (Exception e) {
		}
		this.txtCrossoverProbability.setText(Double.toString(this.solver
				.getCrossoverProbability()));
	}

	/**
	 * Set value of crossover-distribution-index
	 */
	private void setCrossoverDistributionIndex() {
		try {
			this.solver.setCrossoverDistributionIndex(Double
					.parseDouble(this.txtCrossoverDistributionIndex.getText()));
		} catch (Exception e) {
		}
		this.txtCrossoverDistributionIndex.setText(Double.toString(this.solver
				.getCrossoverDistributionIndex()));
	}

	/**
	 * Set value of mutation-probability
	 */
	private void setMutationProbability() {
		try {
			this.solver.setMutationProbability(Double
					.parseDouble(this.txtMutationProbability.getText()));
		} catch (Exception e) {
		}
		this.txtMutationProbability.setText(Double.toString(this.solver
				.getMutationProbability()));
	}

	/**
	 * Set value of mutation-distribution-index
	 */
	private void setMutationDistributionIndex() {
		try {
			this.solver.setMutationDistributionIndex(Double
					.parseDouble(this.txtMutationDistributionIndex.getText()));
		} catch (Exception e) {
		}
		this.txtMutationDistributionIndex.setText(Double.toString(this.solver
				.getMutationDistributionIndex()));
	}

	/**
	 * Set value of DE CR parameter
	 */
	private void setDeCR() {
		try {
			this.solver.setDeCR(Double.parseDouble(this.txtDeCR.getText()));
		} catch (Exception e) {
		}
		this.txtDeCR.setText(Double.toString(this.solver.getDeCR()));
	}

	/**
	 * Set value of DE F parameter
	 */
	private void setDeF() {
		try {
			this.solver.setDeF(Double.parseDouble(this.txtDeF.getText()));
		} catch (Exception e) {
		}
		this.txtDeF.setText(Double.toString(this.solver.getDeF()));
	}

	/**
	 * Initialize components
	 */
	private void initComponents() {
		JLabel labArchiveSize, labArchiveSizeUnit;
		JLabel labCrossoverDistributionIndex, labCrossoverProbability;
		JLabel labMutationDistributionIndex, labMutationProbability;
		JLabel labDeCR, labDeF;

		labCrossoverProbability = new JLabel();
		labCrossoverDistributionIndex = new JLabel();
		labMutationProbability = new JLabel();
		labMutationDistributionIndex = new JLabel();
		labDeCR = new JLabel();
		labDeF = new JLabel();

		labArchiveSize = new JLabel();
		this.txtCrossoverProbability = new JTextField();
		this.txtCrossoverDistributionIndex = new JTextField();
		this.txtMutationProbability = new JTextField();
		this.txtMutationDistributionIndex = new JTextField();
		this.txtArchiveSize = new JTextField();
		this.txtDeCR = new JTextField();
		this.txtDeF = new JTextField();
		labArchiveSizeUnit = new JLabel();

		labCrossoverProbability.setText("Crossover probability:");

		labCrossoverDistributionIndex.setText("Crossover distribution index:");

		labMutationProbability.setText("Mutation probability:");

		labMutationDistributionIndex.setText("Mutation distribution index:");

		labArchiveSize.setText("Archive size:");

		labDeCR.setText("DE CR parameter:");

		labDeF.setText("DE F parameter:");

		this.txtCrossoverProbability.setPreferredSize(new Dimension(100, 20));
		this.txtCrossoverProbability.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaDecmoPanel.this.setCrossoverProbability();
			}
		});
		this.txtCrossoverProbability.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setCrossoverProbability();
			}
		});

		this.txtCrossoverDistributionIndex.setPreferredSize(new Dimension(100,
				20));
		this.txtCrossoverDistributionIndex
				.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent evt) {
						SsaDecmoPanel.this.setCrossoverDistributionIndex();
					}
				});
		this.txtCrossoverDistributionIndex.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setCrossoverDistributionIndex();
			}
		});

		this.txtMutationProbability.setPreferredSize(new Dimension(100, 20));
		this.txtMutationProbability.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaDecmoPanel.this.setMutationProbability();
			}
		});
		this.txtMutationProbability.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setMutationProbability();
			}
		});

		this.txtMutationDistributionIndex.setPreferredSize(new Dimension(100,
				20));
		this.txtMutationDistributionIndex
				.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent evt) {
						SsaDecmoPanel.this.setMutationDistributionIndex();
					}
				});
		this.txtMutationDistributionIndex.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setMutationDistributionIndex();
			}
		});

		this.txtArchiveSize.setPreferredSize(new Dimension(100, 20));
		this.txtArchiveSize.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaDecmoPanel.this.setArchiveSize();
			}
		});
		this.txtArchiveSize.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setArchiveSize();
			}
		});

		labArchiveSizeUnit.setText("Individuals");

		this.txtDeCR.setPreferredSize(new Dimension(100, 20));
		this.txtDeCR.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaDecmoPanel.this.setDeCR();
			}
		});
		this.txtDeCR.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setDeCR();
			}
		});

		this.txtDeF.setPreferredSize(new Dimension(100, 20));
		this.txtDeF.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaDecmoPanel.this.setDeF();
			}
		});
		this.txtDeF.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaDecmoPanel.this.setDeF();
			}
		});

		GroupLayout layout = new GroupLayout(this);
		layout.setHorizontalGroup(layout
				.createParallelGroup(Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												Alignment.LEADING)
												.addComponent(labArchiveSize)
												.addComponent(
														labCrossoverProbability)
												.addComponent(
														labCrossoverDistributionIndex)
												.addComponent(
														labMutationProbability)
												.addComponent(
														labMutationDistributionIndex)
												.addComponent(labDeCR)
												.addComponent(labDeF))
								.addGap(24)
								.addGroup(
										layout.createParallelGroup(
												Alignment.TRAILING, false)
												.addGroup(
														layout.createSequentialGroup()
																.addGroup(
																		layout.createParallelGroup(
																				Alignment.TRAILING)
																				.addComponent(
																						this.txtDeCR,
																						Alignment.LEADING,
																						GroupLayout.DEFAULT_SIZE,
																						60,
																						Short.MAX_VALUE)
																				.addComponent(
																						this.txtMutationDistributionIndex,
																						GroupLayout.DEFAULT_SIZE,
																						60,
																						Short.MAX_VALUE)
																				.addComponent(
																						this.txtDeF,
																						GroupLayout.DEFAULT_SIZE,
																						60,
																						Short.MAX_VALUE))
																.addGap(69))
												.addGroup(
														layout.createSequentialGroup()
																.addGroup(
																		layout.createParallelGroup(
																				Alignment.TRAILING)
																				.addComponent(
																						this.txtCrossoverProbability,
																						GroupLayout.PREFERRED_SIZE,
																						60,
																						GroupLayout.PREFERRED_SIZE)
																				.addComponent(
																						this.txtArchiveSize,
																						GroupLayout.PREFERRED_SIZE,
																						60,
																						GroupLayout.PREFERRED_SIZE)
																				.addComponent(
																						this.txtCrossoverDistributionIndex,
																						GroupLayout.PREFERRED_SIZE,
																						60,
																						GroupLayout.PREFERRED_SIZE)
																				.addComponent(
																						this.txtMutationProbability,
																						GroupLayout.PREFERRED_SIZE,
																						60,
																						GroupLayout.PREFERRED_SIZE))
																.addPreferredGap(
																		ComponentPlacement.UNRELATED)
																.addComponent(
																		labArchiveSizeUnit,
																		GroupLayout.PREFERRED_SIZE,
																		59,
																		GroupLayout.PREFERRED_SIZE)))
								.addContainerGap(149, Short.MAX_VALUE)));
		layout.setVerticalGroup(layout
				.createParallelGroup(Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(labArchiveSize)
												.addComponent(
														this.txtArchiveSize,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(
														labArchiveSizeUnit))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														labCrossoverProbability)
												.addComponent(
														this.txtCrossoverProbability,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														labCrossoverDistributionIndex)
												.addComponent(
														this.txtCrossoverDistributionIndex,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														labMutationProbability)
												.addComponent(
														this.txtMutationProbability,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														labMutationDistributionIndex)
												.addComponent(
														this.txtMutationDistributionIndex,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														this.txtDeCR,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(labDeCR))
								.addPreferredGap(ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												Alignment.BASELINE)
												.addComponent(
														this.txtDeF,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(labDeF))
								.addContainerGap(83, Short.MAX_VALUE)));
		layout.linkSize(SwingConstants.HORIZONTAL, new Component[] {
				this.txtCrossoverProbability,
				this.txtCrossoverDistributionIndex,
				this.txtMutationProbability, this.txtArchiveSize });
		this.setLayout(layout);
	}
}
