/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.me_nsga2;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.JMSolver;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import jmetal.core.SolutionSet;
import jmetal.metaheuristics.me_nsgaII.ME_Nsga2Real;
import jmetal.problems.ME_StandardProblemReal;
import jmetal.util.JMException;

public abstract class ME_Nsga2 extends JMSolver {

	/** Implementation of NSGA-II */
	protected ME_Nsga2Real solver;

	/** Crossover-probability */
	private double crossoverProbability = 1.0;

	/** Mutation-probability */
	private double mutationProbability = 0.01;

	/** Archive-size */
	private int archiveSize = 200;

	@Override
	public String getOptimizerName() {
		return "ME-NSGA-II";
	}

	@Override
	public String getDescription() {
		return "ME-NSGA-II Algorithm, Crossover: SinglePointCrossover, Mutation: BitFlip, selection: BinaryTournament2";
	}

	/**
	 * Constructor of ME-NSGA-II algorithm
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 * @throws JMException
	 */
	public ME_Nsga2(Problem problem) throws JMException {
		super(problem);

		this.solver = new ME_Nsga2Real((ME_StandardProblemReal) this.jmProblem, this.archiveSize,
				this.crossoverProbability, this.mutationProbability);
	}

	@Override
	public TStruct getSettings() throws Exception {
		TStruct resumeData;

		resumeData = new TStruct();

		// archive size
		resumeData.put("archiveSize", new TNumeric(this.getArchiveSize()));

		// crossover probability
		resumeData.put("crossoverProbability", new TNumeric(this.getCrossoverProbability()));

		// mutation probybility
		resumeData.put("mutationProbability", new TNumeric(this.getMutationProbability()));

		return resumeData;
	}

	@Override
	public void setSettings(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive size
		num = (TNumeric) resumeData.get("archiveSize");
		if (num != null) {
			this.setArchiveSize((int) num.getReal(0, 0));
		}

		// crossover probability
		num = (TNumeric) resumeData.get("crossoverProbability");
		if (num != null) {
			this.setCrossoverProbability(num.getReal(0, 0));
		}

		// mutation probability
		num = (TNumeric) resumeData.get("mutationProbability");
		if (num != null) {
			this.setMutationProbability(num.getReal(0, 0));
		}

	}

	@Override
	public TStruct getState() throws Exception {
		TStruct resumeData;
		SolutionSet sol;

		resumeData = new TStruct();

		// archive
		sol = this.solver.getArchive();
		if (sol.size() != 0) {
			resumeData.put("archive", new TNumeric(Archive
					.int2double(this.getProblem().getArchive().getMemory().getIndexOf(this.getIndividuals(sol)))));
		}

		return resumeData;
	}

	@Override
	public void setState(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive
		num = (TNumeric) resumeData.get("archive");
		if (num != null) {
			this.solver.setArchive(this.getSolutionSet(this.getProblem().getArchive().getMemory()
					.getIndividual(Archive.double2int(num.getRealMatrix().getRow(0)))));
		}
	}

	@Override
	public Individual[] getArchive() throws Exception {
		return this.getIndividuals(this.solver.getArchive());
	}

	@Override
	public void setArchive(Individual[] ind) throws Exception {
		this.solver.setArchive(this.getSolutionSet(ind));

	}

	@Override
	public Individual[] getPareto() throws Exception {
		return this.getIndividuals(this.solver.getPareto());
	}

	public double getCrossoverProbability() {
		return this.crossoverProbability;
	}

	public void setCrossoverProbability(double crossoverProbability) {
		this.crossoverProbability = crossoverProbability;
	}

	public double getMutationProbability() {
		return this.mutationProbability;
	}

	public void setMutationProbability(double mutationProbability) {
		this.mutationProbability = mutationProbability;
	}

	public int getArchiveSize() {
		return this.archiveSize;
	}

	public void setArchiveSize(int archiveSize) {
		this.archiveSize = archiveSize;
	}
}
