/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.nsga2;

import jmetal.core.SolutionSet;
import jmetal.metaheuristics.nsgaII.Nsga2Real;
import jmetal.util.JMException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.JMSolver;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;

/**
 * NSGA-II implementation
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public abstract class Nsga2 extends JMSolver {

	/** Implementation of NSGA-II */
	protected Nsga2Real solver;

	/** Crossover-probability */
	private double crossoverProbability = 0.9;

	/** Crossover-distribution-index */
	private double crossoverDistributionIndex = 20;

	/** Mutation-probability */
	private double mutationProbability = 0.4;

	/** Mutation-distribution-index */
	private double mutationDistributionIndex = 5;

	/** Archive-size */
	private int archiveSize = 200;

	/**
	 * Constructor of NSGA-II algorithm
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 * @throws JMException
	 */
	public Nsga2(Problem problem) throws JMException {
		super(problem);

		this.solver = new Nsga2Real(this.jmProblem, this.archiveSize,
				this.crossoverProbability, this.crossoverDistributionIndex,
				this.mutationProbability, this.mutationDistributionIndex);
	}

	@Override
	public TStruct getSettings() throws Exception {
		TStruct resumeData;

		resumeData = new TStruct();

		// archive size
		resumeData.put("archiveSize", new TNumeric(this.getArchiveSize()));

		// crossover probability
		resumeData.put("crossoverProbability",
				new TNumeric(this.getCrossoverProbability()));

		// crossover distribution index
		resumeData.put("crossoverDistributionIndex",
				new TNumeric(this.getCrossoverDistributionIndex()));

		// mutation probybility
		resumeData.put("mutationProbability",
				new TNumeric(this.getMutationProbability()));

		// mutation distribution index
		resumeData.put("mutationDistributionIndex",
				new TNumeric(this.getMutationDistributionIndex()));

		return resumeData;
	}

	@Override
	public void setSettings(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive size
		num = (TNumeric) resumeData.get("archiveSize");
		if (num != null) {
			this.setArchiveSize((int) num.getReal(0, 0));
		}

		// crossover probability
		num = (TNumeric) resumeData.get("crossoverProbability");
		if (num != null) {
			this.setCrossoverProbability(num.getReal(0, 0));
		}

		// crossover distribution index
		num = (TNumeric) resumeData.get("crossoverDistributionIndex");
		if (num != null) {
			this.setCrossoverDistributionIndex(num.getReal(0, 0));
		}

		// mutation probability
		num = (TNumeric) resumeData.get("mutationProbability");
		if (num != null) {
			this.setMutationProbability(num.getReal(0, 0));
		}

		// mutation distribution index
		num = (TNumeric) resumeData.get("mutationDistributionIndex");
		if (num != null) {
			this.setMutationDistributionIndex(num.getReal(0, 0));
		}
	}

	@Override
	public TStruct getState() throws Exception {
		TStruct resumeData;
		SolutionSet sol;

		resumeData = new TStruct();

		// archive
		sol = this.solver.getArchive();
		if (sol.size() != 0) {
			resumeData.put(
					"archive",
					new TNumeric(Archive.int2double(this.getProblem()
							.getArchive().getMemory()
							.getIndexOf(this.getIndividuals(sol)))));
		}

		return resumeData;
	}

	@Override
	public void setState(TStruct resumeData) throws Exception {
		TNumeric num;

		// archive
		num = (TNumeric) resumeData.get("archive");
		if (num != null) {
			this.solver
					.setArchive(this.getSolutionSet(this
							.getProblem()
							.getArchive()
							.getMemory()
							.getIndividual(
									Archive.double2int(num.getRealMatrix()
											.getRow(0)))));
		}
	}

	@Override
	public String getOptimizerName() {
		return "Nsga2";
	}

	@Override
	public String getDescription() {
		return "NSGA-II Algorithm, Crossover: SBX, Mutation: polynomial, selection: BinaryTournament2";
	}

	/**
	 * Get crossover-probability
	 * 
	 * @return Value
	 */
	public double getCrossoverProbability() {
		return this.crossoverProbability;
	}

	/**
	 * Get size of archive
	 * 
	 * @return Size of archive
	 */
	public int getArchiveSize() {
		return this.archiveSize;
	}

	/**
	 * Set size of archive
	 * 
	 * @param archiveSize
	 *            Size of archive
	 */
	public void setArchiveSize(int archiveSize) {
		this.archiveSize = archiveSize;
	}

	/**
	 * Set crossover-probability
	 * 
	 * @param crossoverProbability
	 *            Value
	 */
	public void setCrossoverProbability(double crossoverProbability) {
		this.crossoverProbability = crossoverProbability;
	}

	/**
	 * Get crossover-distribution-index
	 * 
	 * @return Value
	 */
	public double getCrossoverDistributionIndex() {
		return this.crossoverDistributionIndex;
	}

	/**
	 * Set crossover-distribution-index
	 * 
	 * @param crossoverDistributionIndex
	 *            Value
	 */
	public void setCrossoverDistributionIndex(double crossoverDistributionIndex) {
		this.crossoverDistributionIndex = crossoverDistributionIndex;
	}

	/**
	 * Get mutation-probability
	 * 
	 * @return Value
	 */
	public double getMutationProbability() {
		return this.mutationProbability;
	}

	/**
	 * Set mutation-probability
	 * 
	 * @param mutationProbability
	 *            Value
	 */
	public void setMutationProbability(double mutationProbability) {
		this.mutationProbability = mutationProbability;
	}

	/**
	 * Get mutation-distribution-index
	 * 
	 * @return Value
	 */
	public double getMutationDistributionIndex() {
		return this.mutationDistributionIndex;
	}

	/**
	 * Set mutation-distribution-index
	 * 
	 * @param mutationDistributionIndex
	 *            Value
	 */
	public void setMutationDistributionIndex(double mutationDistributionIndex) {
		this.mutationDistributionIndex = mutationDistributionIndex;
	}

	@Override
	public Individual[] getArchive() throws Exception {
		return this.getIndividuals(this.solver.getArchive());
	}

	@Override
	public void setArchive(Individual[] ind) throws Exception {
		this.solver.setArchive(this.getSolutionSet(ind));
	}

	@Override
	public Individual[] getPareto() throws Exception {
		return this.getIndividuals(this.solver.getPareto());
	}
}
