/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.spea2;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;

import javax.swing.GroupLayout;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.SwingConstants;

import at.lcm.bu21.optim.algorithm.JPanelSolver;

/**
 * Panel for settings of spea2real optimizer
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public class SsaSpea2Panel extends JPanelSolver {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Spea2Real optimizer with setting values */
	private Spea2 solver;

	/** Size of archive */
	private JTextField txtArchiveSize;

	/** Crossover distribution index */
	private JTextField txtCrossoverDistributionIndex;

	/** Crossover probability */
	private JTextField txtCrossoverProbability;

	/** Mutation distribution index */
	private JTextField txtMutationDistributionIndex;

	/** Mutation probability */
	private JTextField txtMutationProbability;

	/**
	 * Create panel. Used by WindowsBuilder
	 */
	private SsaSpea2Panel() {
		this.initComponents();
	}

	/**
	 * Creates new setting-panel for Spea2Real
	 * 
	 * @param solver
	 *            Optimizer with setting values
	 */
	public SsaSpea2Panel(Spea2 solver) {
		this();
		this.solver = solver;

		this.txtArchiveSize.setText(Integer.toString(this.solver
				.getArchiveSize()));
		this.txtCrossoverProbability.setText(Double.toString(this.solver
				.getCrossoverProbability()));
		this.txtCrossoverDistributionIndex.setText(Double.toString(this.solver
				.getCrossoverDistributionIndex()));
		this.txtMutationProbability.setText(Double.toString(this.solver
				.getMutationProbability()));
		this.txtMutationDistributionIndex.setText(Double.toString(this.solver
				.getMutationDistributionIndex()));

		this.refreshEnabled();
	}

	@Override
	public void refreshEnabled() {
		boolean en;

		en = this.enable && !this.solver.getProblem().isLocked();

		this.txtArchiveSize.setEnabled(en);
		this.txtCrossoverProbability.setEnabled(en);
		this.txtCrossoverDistributionIndex.setEnabled(en);
		this.txtMutationProbability.setEnabled(en);
		this.txtMutationDistributionIndex.setEnabled(en);
	}

	/**
	 * Set value of archive-size
	 */
	private void setArchiveSize() {
		try {
			this.solver.setArchiveSize(Integer.parseInt(this.txtArchiveSize
					.getText()));
		} catch (Exception e) {
		}
		this.txtArchiveSize.setText(Integer.toString(this.solver
				.getArchiveSize()));
	}

	/**
	 * Set value of crossover-probability
	 */
	private void setCrossoverProbability() {
		try {
			this.solver.setCrossoverProbability(Double
					.parseDouble(this.txtCrossoverProbability.getText()));
		} catch (Exception e) {
		}
		this.txtCrossoverProbability.setText(Double.toString(this.solver
				.getCrossoverProbability()));
	}

	/**
	 * Set value of crossover-distribution-index
	 */
	private void setCrossoverDistributionIndex() {
		try {
			this.solver.setCrossoverDistributionIndex(Double
					.parseDouble(this.txtCrossoverDistributionIndex.getText()));
		} catch (Exception e) {
		}
		this.txtCrossoverDistributionIndex.setText(Double.toString(this.solver
				.getCrossoverDistributionIndex()));
	}

	/**
	 * Set value of mutation-probability
	 */
	private void setMutationProbability() {
		try {
			this.solver.setMutationProbability(Double
					.parseDouble(this.txtMutationProbability.getText()));
		} catch (Exception e) {
		}
		this.txtMutationProbability.setText(Double.toString(this.solver
				.getMutationProbability()));
	}

	/**
	 * Set value of mutation-distribution-index
	 */
	private void setMutationDistributionIndex() {
		try {
			this.solver.setMutationDistributionIndex(Double
					.parseDouble(this.txtMutationDistributionIndex.getText()));
		} catch (Exception e) {
		}
		this.txtMutationDistributionIndex.setText(Double.toString(this.solver
				.getMutationDistributionIndex()));
	}

	/**
	 * Initialize components
	 */
	private void initComponents() {
		JLabel labArchiveSize, labArchiveSizeUnit;
		JLabel labCrossoverDistributionIndex, labCrossoverProbability;
		JLabel labMutationDistributionIndex, labMutationProbability;

		labCrossoverProbability = new JLabel();
		labCrossoverDistributionIndex = new JLabel();
		labMutationProbability = new JLabel();
		labMutationDistributionIndex = new JLabel();
		labArchiveSize = new JLabel();
		this.txtCrossoverProbability = new JTextField();
		this.txtCrossoverDistributionIndex = new JTextField();
		this.txtMutationProbability = new JTextField();
		this.txtMutationDistributionIndex = new JTextField();
		this.txtArchiveSize = new JTextField();
		labArchiveSizeUnit = new JLabel();

		labCrossoverProbability.setText("Crossover probability:");

		labCrossoverDistributionIndex.setText("Crossover distribution index:");

		labMutationProbability.setText("Mutation probability:");

		labMutationDistributionIndex.setText("Mutation distribution index:");

		labArchiveSize.setText("Archive size:");

		this.txtCrossoverProbability.setPreferredSize(new Dimension(100, 20));
		this.txtCrossoverProbability.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaSpea2Panel.this.setCrossoverProbability();
			}
		});
		this.txtCrossoverProbability.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaSpea2Panel.this.setCrossoverProbability();
			}
		});

		this.txtCrossoverDistributionIndex.setPreferredSize(new Dimension(100,
				20));
		this.txtCrossoverDistributionIndex
				.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent evt) {
						SsaSpea2Panel.this.setCrossoverDistributionIndex();
					}
				});
		this.txtCrossoverDistributionIndex.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaSpea2Panel.this.setCrossoverDistributionIndex();
			}
		});

		this.txtMutationProbability.setPreferredSize(new Dimension(100, 20));
		this.txtMutationProbability.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaSpea2Panel.this.setMutationProbability();
			}
		});
		this.txtMutationProbability.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaSpea2Panel.this.setMutationProbability();
			}
		});

		this.txtMutationDistributionIndex.setPreferredSize(new Dimension(100,
				20));
		this.txtMutationDistributionIndex
				.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent evt) {
						SsaSpea2Panel.this.setMutationDistributionIndex();
					}
				});
		this.txtMutationDistributionIndex.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaSpea2Panel.this.setMutationDistributionIndex();
			}
		});

		this.txtArchiveSize.setPreferredSize(new Dimension(100, 20));
		this.txtArchiveSize.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent evt) {
				SsaSpea2Panel.this.setArchiveSize();
			}
		});
		this.txtArchiveSize.addFocusListener(new FocusAdapter() {
			@Override
			public void focusLost(FocusEvent evt) {
				SsaSpea2Panel.this.setArchiveSize();
			}
		});

		labArchiveSizeUnit.setText("Individuals");

		GroupLayout layout = new GroupLayout(this);
		this.setLayout(layout);
		layout.setHorizontalGroup(layout
				.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.LEADING)
												.addComponent(labArchiveSize)
												.addComponent(
														labCrossoverProbability)
												.addComponent(
														labCrossoverDistributionIndex)
												.addComponent(
														labMutationProbability)
												.addComponent(
														labMutationDistributionIndex))
								.addGap(24, 24, 24)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.LEADING)
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		this.txtMutationDistributionIndex,
																		GroupLayout.PREFERRED_SIZE,
																		60,
																		GroupLayout.PREFERRED_SIZE)
																.addContainerGap(
																		237,
																		Short.MAX_VALUE))
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		this.txtMutationProbability,
																		GroupLayout.PREFERRED_SIZE,
																		60,
																		GroupLayout.PREFERRED_SIZE)
																.addContainerGap(
																		237,
																		Short.MAX_VALUE))
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		this.txtCrossoverDistributionIndex,
																		0,
																		60,
																		Short.MAX_VALUE)
																.addGap(237,
																		237,
																		237))
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		this.txtArchiveSize,
																		GroupLayout.PREFERRED_SIZE,
																		60,
																		GroupLayout.PREFERRED_SIZE)
																.addPreferredGap(
																		LayoutStyle.ComponentPlacement.UNRELATED)
																.addComponent(
																		labArchiveSizeUnit,
																		GroupLayout.PREFERRED_SIZE,
																		59,
																		GroupLayout.PREFERRED_SIZE)
																.addContainerGap())
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		this.txtCrossoverProbability,
																		GroupLayout.PREFERRED_SIZE,
																		60,
																		GroupLayout.PREFERRED_SIZE)
																.addContainerGap(
																		237,
																		Short.MAX_VALUE)))));

		layout.linkSize(SwingConstants.HORIZONTAL,
				new Component[] { this.txtArchiveSize,
						this.txtCrossoverDistributionIndex,
						this.txtCrossoverProbability,
						this.txtMutationDistributionIndex,
						this.txtMutationProbability });

		layout.setVerticalGroup(layout
				.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(labArchiveSize)
												.addComponent(
														this.txtArchiveSize,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE)
												.addComponent(
														labArchiveSizeUnit))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labCrossoverProbability)
												.addComponent(
														this.txtCrossoverProbability,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labCrossoverDistributionIndex)
												.addComponent(
														this.txtCrossoverDistributionIndex,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labMutationProbability)
												.addComponent(
														this.txtMutationProbability,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(
										LayoutStyle.ComponentPlacement.UNRELATED)
								.addGroup(
										layout.createParallelGroup(
												GroupLayout.Alignment.BASELINE)
												.addComponent(
														labMutationDistributionIndex)
												.addComponent(
														this.txtMutationDistributionIndex,
														GroupLayout.PREFERRED_SIZE,
														GroupLayout.DEFAULT_SIZE,
														GroupLayout.PREFERRED_SIZE))
								.addContainerGap(144, Short.MAX_VALUE)));
	}
}
