#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9967 $
$LastChangedDate:: 2020-08-24 11:17:15 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import uuid
import glob
import optparse
import tempfile
import shutil
import datetime
import time
import xml.etree.ElementTree as ET
import symspace.osutils.processutils as pu
import symspace.osutils.fileutils as fu
from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser
from symspace.testing.junit2html import junit2html, junit2htmlMatrix
from symspace.testing.moptestrunner import SymProjectTestsuite


class SymUnitTestRunner(SymProjectTestsuite):
    """Tests all SyMSpace projects (with file extension *.mop) found in path and all it's subdirectories

    Adds currently no additional features to SymProjectTestsuite
    """

    def __init__(self, path, run_args=None):
        super().__init__(path, run_args=run_args)


class SymXmlTestRunner:

    def __init__(self, start_dir: str, reports_dir: str, outsuffix: str = None, run_args=None, timeout: float = 60):
        """
        Tests SyMSpace projects (with file extension *.mop) found in path and all it's subdirectories
        and writes xml result files

        Parameters
        ----------
        start_dir : str
            Start path for test collection, all subdirectories are included
        reports_dir: str
            Output directory for the xml testreport
        outsuffix : str
            Suffix for the xml and testsuit name, None (Default)

            filename  = 'symspace.moptestrunner.SymProjectTestCase' + outsuffix + '.xml'
        run_args: dict
            SyMSpace run arguments
        timeout : float
            Timeout (in seconds) for single test

        """
        self.start_dir = start_dir
        self.reports_dir = reports_dir
        self.junit_file = os.path.join(self.reports_dir, 'test_results-junit.xml')
        self.outsuffix = outsuffix
        self.search_pattern = '\\**\\test_*.mop'
        if run_args is None:
            run_args = dict()
        self.run_args = run_args
        self.timeout = timeout
        self.suppress_output = False
        self.detailed_error = False

        if os.path.isdir(self.reports_dir):
            shutil.rmtree(self.reports_dir, onerror=fu.on_rm_error)
        os.makedirs(self.reports_dir, exist_ok=True)

    @staticmethod
    def discover(dirpath: str, search_pattern: str = None) -> list:
        """
        Discover testable .mop files in dirpath.

        Parameters
        ----------
        dirpath : str
            Path to directory
        search_pattern : str
            Additional search pattern, None (Default)

        Returns
        -------
        ret : list
            List of full path names

        """
        if search_pattern is None:
            files = SymProjectTestsuite.discover(dirpath)
        else:
            files = glob.glob(dirpath + search_pattern, recursive=True)
        return files

    def run(self):
        """
        Run all tests for the given search pattern, beginning with the start directory
        """
        print('Start Directory: {}'.format(self.start_dir))

        files = self.discover(self.start_dir, self.search_pattern)
        self.runTests(files)

    def runTests(self, files: list):
        """
        Run tests on the given files

        Parameters
        ----------
        files : list
            List of files to test
        """
        tmp_dir = tempfile.mkdtemp()
        n_test = len(files)

        print('\nFound {} SyMSpace tests'.format(n_test))

        for cnt_tests, path in enumerate(files):
            disp_path = path.replace(Paths.getComponentSpaceRoot(), '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Run Test {}/{} ==> {}'.format(timestamp, cnt_tests + 1, n_test, disp_path))
            try:
                self.runScript(path, tmp_dir)
            except TimeoutError as ext:
                print('        ==> Failed: {}'.format(ext))
            except Exception as ex:
                if self.detailed_error:
                    print('        ==> Failed: {}'.format(ex))
                else:
                    print('        ==> Failed')

        self.mergeJUnitFiles(tmp_dir)

        try:
            print('Clean up temporary files')
            time.sleep(2)
            shutil.rmtree(tmp_dir)
        except Exception as ex:
            print(ex)

    def runScript(self, path: str, tmp_dir: str):
        script = """
import xmlrunner
from symspace.testing.moptestrunner import SymProjectTestsuite
run_args = {run_args}
test_suite = SymProjectTestsuite("{path}", run_args)
test_suite.runner = xmlrunner.XMLTestRunner(outsuffix="{outsuffix}", output="{reports_dir}")
result = test_suite.run()
if len(result.errors) > 0 or len(result.failures) > 0:
    raise Exception('Failed with errors')
        """
        file_id = uuid.uuid4()
        outsuffix = '{}-{}-junit'.format(self.outsuffix, file_id)
        junit_xml_wildcard = os.path.join(tmp_dir, f'*{outsuffix}.xml')
        script_py = os.path.join(tmp_dir, 'script-{}.py'.format(file_id))

        with open(script_py, 'w') as fh:
            fh.write(script.format(path=path.replace('\\', '\\\\'),
                                   outsuffix=outsuffix,
                                   reports_dir=tmp_dir.replace('\\', '\\\\'),
                                   run_args=self.run_args))

        if self.suppress_output:
            pu.csymspaceRun(script_py, stdout=None, stderr=None, timeout=self.timeout)
        else:
            pu.csymspaceRun(script_py, timeout=self.timeout)

        if len(glob.glob(junit_xml_wildcard)) < 1:
            raise FileNotFoundError('JUnit file not found!')

    def mergeJUnitFiles(self, tmp_dir: str):
        """
        Merge specific content of all junit.xml files in the given directory into one junit-xml

        Parameters
        ----------
        tmp_dir : str
            Search directory
        """
        failures = 0
        tests = 0
        errors = 0
        dt = 0.0
        cases = []

        xml_files = glob.glob(os.path.join(tmp_dir, '*-junit.xml'))

        for file_name in xml_files:
            tree = ET.parse(file_name)
            test_suite = tree.getroot()
            failures += int(test_suite.attrib['failures'])
            tests += int(test_suite.attrib['tests'])
            errors += int(test_suite.attrib['errors'])
            dt += float(test_suite.attrib['time'])

            case = list(test_suite.iter('testcase'))[0]
            sys_out = list(test_suite.iter('system-out'))[0]
            sys_err = list(test_suite.iter('system-err'))[0]

            case.extend([sys_out, sys_err])
            cases.append(case)

        new_root = ET.Element('testsuite')
        new_root.attrib['failures'] = str(failures)
        new_root.attrib['tests'] = str(tests)
        new_root.attrib['errors'] = str(errors)
        new_root.attrib['time'] = str(dt)

        for case in cases:
            new_root.append(case)

        new_tree = ET.ElementTree(new_root)
        new_tree.write(self.junit_file)
        print('File saved to: {}'.format(self.junit_file))

    @classmethod
    def fromCMDParser(cls, default_repo='SyMSpace') -> 'SymXmlTestRunner':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --startDir : str
            The search for component starts from this directory (Optional).
            Default is the SyMSpace root directory.

        Returns
        -------
        runner : SymXmlTestRunner
        """

        if default_repo == 'SyMSpace':
            default_start_dir = Paths.getSyMSpaceRoot()
        elif default_repo == 'ComponentSpace':
            default_start_dir = Paths.getComponentSpaceRoot()
        else:
            raise Exception('Unknown default repo {}'.format(default_repo))

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-r", "--repo", dest="repository",
                          default=default_repo, type="string",
                          help="Repository")
        parser.add_option("-d", "--startDir", dest="start_dir", default=default_start_dir,
                          type="string", help="The search for component starts from this directory, "
                                              "default is the SyMSpace root directory.")
        parser.add_option("--runSim", dest="RunSim", default=0,
                          type="int", help="Set run_args RunSim")
        parser.add_option("--remoteCompute", dest="RemoteCompute", default=0,
                          type="int", help="Set run_args RemoteCompute")
        parser.add_option("--optimize", dest="Optimize", default=0,
                          type="int", help="Set run_args Optimize")
        parser.add_option("--timeout", dest="timeout", default=60000,
                          type="int", help="Timeout (in milliseconds) for single test")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)

        if len(args) != 0:
            parser.error("incorrect number of arguments")

        repo = options.repository
        start_dir = CMDParser.jepArgPath(options.start_dir)

        run_sim = bool(options.RunSim)
        remote_compute = bool(options.RemoteCompute)
        optimize = bool(options.Optimize)
        timeout = float(options.timeout / 1000)  # to seconds

        if repo == 'SyMSpace':
            suffix = os.path.normpath(start_dir).replace(Paths.getSyMSpaceRoot(), '').replace(os.path.sep, '.')
            outsuffix = 'SyMSpace{}'.format(suffix)
            output_dir = os.path.join(Paths.getCSTestReportsDir(), 'run_project_tests')
        elif repo == 'ComponentSpace':
            suffix = os.path.normpath(start_dir).replace(Paths.getComponentSpaceRoot(), '').replace(os.path.sep, '.')
            outsuffix = 'ComponentSpace{}'.format(suffix)
            output_dir = os.path.join(Paths.getCSTestReportsDir(), 'run_project_tests')
        else:
            raise Exception('Unknown repo {}'.format(repo))

        print('Start Path: {}'.format(start_dir))
        print('Single Test Timeout: {:.1f} seconds'.format(timeout))
        print('RunSim: {}, RemoteCompute: {}, Optimize: {}'.format(run_sim, remote_compute, optimize))

        run_args = dict(RunSim=run_sim, RemoteCompute=remote_compute, Optimize=optimize)
        runner = cls(start_dir, reports_dir=output_dir, outsuffix=outsuffix, run_args=run_args, timeout=timeout)
        return runner

    def getHTMLReport(self, matrix: bool = True, show: bool = False) -> str:
        """
        Convert the junit report into html

        Parameters
        ----------
        matrix : bool
            Run junit2htmlMatrix or junit2html
        show : bool
            Open in webbrowser

        Returns
        -------
        html_file : str
            Path to the created html file
        """
        if matrix:
            html_file = junit2htmlMatrix(self.junit_file, show=show)
        else:
            html_file = junit2html(self.junit_file, show=show)

        return html_file


# class SymNoseTestRunner(SymProjectTestsuite):
#
#     def __init__(self, path, output_dir):
#         import nose
#         super().__init__(path)
#         self.output_dir = output_dir
#         self.runner = nose
#
#     def run(self):
#         """
#         Function for running the created testsuite
#         """
#         xmlfile = os.path.join(self.output_dir, 'nosetests-junit.xml')
#
#         if not os.path.isdir(self.output_dir):
#             os.makedirs(self.output_dir)
#
#         argv = ['SymProjects',
#                 '--with-xunit', '--xunit-file={}'.format(xmlfile)]
#         return self.runner.run(argv=argv, suite=self.suite) # xunit does not support suites!!!!


if __name__ == "__main__":
    start_path = os.path.join(os.getenv('ComponentSpace'), 'Part', 'tests')

    # print('\n################# UnitTestRunner ##############')
    # unitTestRunner = SymUnitTestRunner(start_path)
    # result1 = unitTestRunner.run()

    print('\n################# XMLTestRunner ##############')
    xmlTestRunner = SymXmlTestRunner(start_path, 'x:\\xml_testrunner')
    xmlTestRunner.run()

    # print('\n################# NoseTestRunner ##############')
    # noseTestRunner = SymNoseTestRunner(start_path, 'x:\\nose_testrunner')
    # result3 = noseTestRunner.run()
