# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10096 $
$LastChangedDate:: 2021-01-21 16:08:58 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import pandas as pd
from symspace.wrapper.symoptwrapper import SymArchive, SymProblem, SymIndividual, SymMemory


class SyMOptUtils:
    """Helper class for |sym| Optimization"""

    @staticmethod
    def getMapShortToFullIds(archive: SymArchive) -> dict:
        """
        Get the mapping of short parameter ids to full ids

        This is a helper method to convert dataframes exported with XLSWriter class


        Parameters
        ----------
        archive :SymArchive

        Returns
        -------
        short2full : dict
            mapping of short parameter ids (=key) to full ids (=value)

        """
        prj_ids = archive.getVariables()
        prj_ids.extend(archive.getResults())
        project = archive.getProject()

        short2full = dict()
        for full_id in prj_ids:
            param = project[full_id]
            short_id = param.getShortIdentifier()
            if short_id in short2full:
                raise Exception('Mapping short to full ids failed!')

            short2full[short_id] = full_id

        return short2full

    @classmethod
    def renameDataFrame(cls, archive, df: pd.DataFrame, map_ids: bool) -> pd.DataFrame:
        """
        Rename columns of a dataframe

        This is a helper method to convert dataframes exported with XLSWriter class

        Parameters
        ----------
        archive : SymArchive
        df: pd.DataFrame
        map_ids: bool
            Map short ids (column names ) to full ids

        Returns
        -------

        """
        rename_objectives = dict()
        for key in df.keys():
            rename_objectives[key] = key.replace(' (MAX)', '').replace(' (MIN)', '')

        df.rename(columns=rename_objectives, inplace=True)

        if map_ids:
            short2full = SyMOptUtils.getMapShortToFullIds(archive)
            df.rename(columns=short2full, inplace=True)
        return df

    @classmethod
    def mergeArchives(cls, archives: list, new_name: str) -> SymArchive:
        """
        Merge individuals of multiple archives into a new archive

        Parameters
        ----------
        archives : list
            List of archives to merge
        new_name: str
            Name of the new archive

        Returns
        -------
        new_archive: SymArchive
            New archive
        """
        new_archive = archives[0].duplicate(new_name)

        for archive in archives:
            for idv in archive.getMemory().getIndividuals():
                new_idv = SymIndividual.new(new_archive)  # type: SymIndividual

                for var_name in new_archive.getVariables():
                    value = idv.getVariable(var_name)
                    new_idv.setVariable(var_name, value)
                for result_name in new_archive.getResults():
                    value = idv.getResult(result_name)
                    new_idv.setResult(result_name, value)

                    new_archive.getMemory().addIndividual(new_idv)
        new_archive.save()
        return new_archive

    @staticmethod
    def addIndividuals2Archive(archive: SymArchive, df: pd.DataFrame):
        """
        Add individuals to a archive

        Parameters
        ----------
        archive : SymArchive
        df : pd.DataFrame
        """

        for _, row in df.iterrows():
            new_idv = SymIndividual.new(archive)  # type: SymIndividual

            for var_name in archive.getVariables():
                if var_name not in row:
                    raise KeyError('Variable column "{}" is missing in the DataFrame'.format(var_name))
                value = row[var_name]
                new_idv.setVariable(var_name, value)
            for result_name in archive.getResults():
                if result_name not in row:
                    raise KeyError('Result column "{}" is missing in the DataFrame'.format(result_name))
                value = row[result_name]
                new_idv.setResult(result_name, value)

            archive.getMemory().addIndividual(new_idv)
        archive.save()

    @staticmethod
    def getIndividualsFromType(problem: SymProblem, individual_type: str) -> list:
        """
        Get individuals to the selected type from the memory

        Parameters
        ----------
        problem: SymProblem
            Current problem
        individual_type: str
            Individual type to export from the other problem
            One of: 'all', 'valid', 'no_error', 'error', 'pareto'

        Returns
        -------
        individuals : list
        """
        archive = problem.getArchive()
        memory = archive.getMemory()

        ind_type = individual_type.lower()

        if ind_type == 'no_error':
            individuals = memory.getNoErrorIndividuals()
        elif ind_type == 'valid':
            individuals = memory.getValidIndividuals(problem)
        elif ind_type == 'pareto':
            individuals = memory.getParetoIndividuals(problem)
        elif ind_type == 'error':
            individuals = memory.getErrorIndividuals()
        elif ind_type == 'all':
            individuals = memory.getIndividuals()
        else:
            raise Exception('Unknown option: {}'.format(individual_type))

        return individuals

    @classmethod
    def addExistingIndividuals2Buffer(cls, problem: SymProblem, problem_other: SymProblem, individual_type: str):
        """
        Add individuals of a existing optimization problem to the current optimization buffer

        Parameters
        ----------
        problem :SymProblem
            Current problem
        problem_other: SymProblem
            Other problem
        individual_type: str
            Individual type to export from the other problem
            One of: 'all', 'valid', 'no_error', 'error', 'pareto'
        """
        archive = problem.getArchive()

        individuals_other = cls.getIndividualsFromType(problem_other, individual_type)

        if len(individuals_other) > 0:
            new_individuals = cls.getNewIndividuals(archive, individuals_other)

            cls.addValidVarIndividuals2Buffer(problem, new_individuals)
        else:
            print('No individuals of type "{}" found in the selected archive/problem'.format(individual_type))

    @classmethod
    def addCustomIndividuals2Buffer(cls, problem: SymProblem, df_individuals: pd.DataFrame):
        """
        Create custom individuals and add them to the current optimization buffer

        Parameters
        ----------
        problem: SymProblem
            Current problem
        df_individuals : pd.DataFrame
            DataFrame with optimization Variables and the values for each new individual
        """
        archive = problem.getArchive()
        var_names = archive.getVariables()

        var_names_diff = set(var_names).difference(list(df_individuals.keys()))
        if len(var_names_diff) > 0:
            raise Exception('Required variables are missing: {}'.format(var_names_diff))

        n_idv = df_individuals.shape[0]
        print('Trying to create individuals:')
        new_individuals = [SymIndividual.new(archive) for _ in range(n_idv)]

        for index, row in df_individuals.iterrows():
            for var_name in var_names:
                new_individuals[index].setVariable(var_name, row[var_name])

        cls.addValidVarIndividuals2Buffer(problem, new_individuals)

    @staticmethod
    def addValidVarIndividuals2Buffer(problem: SymProblem, new_individuals: list):
        """
        Add individuals with valid variables to the current optimization buffer

        Parameters
        ----------
        problem: SymProblem
            Current problem
        new_individuals: list[SymIndividuals]
            List of individuals to add
        """
        n_idv = len(new_individuals)
        for idv in new_individuals:
            if not idv.isValidVar(problem):
                print('Invalid Individual:\n{}'.format(idv))
                new_individuals.remove(idv)
        n_idv_valid_var = len(new_individuals)

        print('Adding {:d}/{:d} individuals to buffer'.format(n_idv_valid_var, n_idv))
        opt = problem.getOptimization()
        opt.add2Buffer(new_individuals)

    @staticmethod
    def getNewIndividuals(archive: SymArchive, individuals_other: list) -> list:
        """
        Create new individuals for the current archive from a list of given individuals

        Parameters
        ----------
        archive: SymArchive
            Current archive
        individuals_other: list[SymIndividuals]
            List of individuals from another archive

        Returns
        -------

        new_individuals: list[SymIndividuals]
            List of individuals for the current archive
        """
        n_idv = len(individuals_other)
        new_individuals = [SymIndividual.new(archive) for _ in range(n_idv)]

        if n_idv > 0:
            var_names = archive.getVariables()
            archive_other = individuals_other[0].getArchive()  # all individuals should be of the same archive
            var_names_other = archive_other.getVariables()

            missing_vars = set(var_names).difference(var_names_other)
            if len(missing_vars) > 0:
                '''check if all necessary variables for the current archive are available in the other archive'''
                raise Exception('Optimization variables are missing in the current archive: {}'.format(missing_vars))

            for idv_idx, idv_other in enumerate(individuals_other):
                for var_name in var_names:
                    value = idv_other.getVariable(var_name)
                    new_individuals[idv_idx].setVariable(var_name, value)

        return new_individuals

    @classmethod
    def individualsToDataFrame(cls, problem: SymProblem, individual_type: str) -> pd.DataFrame:
        """
        Get individuals to as DataFrame

        Parameters
        ----------
        problem: SymProblem
            Other problem
        individual_type: str
            Individual type to export from the other problem
            One of: 'all', 'valid', 'no_error', 'error', 'pareto'

        Returns
        -------
        df : pd.DataFrame
            DataFrame of individuals
        """
        individuals = cls.getIndividualsFromType(problem, individual_type)

        df = pd.DataFrame()
        for idv in individuals:
            row = pd.DataFrame.from_dict(idv.toDict(), orient='index').transpose()
            df = df.append(row, ignore_index=True)

        return df
