# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10009 $
$LastChangedDate:: 2020-09-30 13:44:09 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import time
import unittest
from symspace.osutils.processutils import getInterpreter
from symspace.osutils.pathutils import Paths
from symspace.wrapper.symwrapper import SymProject, SymSimulateAll


class TestJobinterface(unittest.TestCase):
    """Test if the Python Jobinterface works correctly"""

    @staticmethod
    def run_sim(sym_file: str) -> tuple:
        """
        Run Simulation helper function

        Parameters
        ----------
        sym_file : str
            SyMSpace Project file to load

        Returns
        -------
        tuple
            status : str
                If status is non-zero the regeneration was incomplete or an error occured
            event_log : str
                Event log of the project
        """
        sym_prj = SymProject.load(sym_file)
        sym_prj.resetAllFunctions()
        sym_prj.getEventLog()

        sim_all = SymSimulateAll(sym_prj)
        sim_all.setRunSimulation(True)
        sim_all.setRemoteCompute(False)
        sim_all.setOptimize(False)
        sim_all.start()

        while sim_all.getNumNotInitialized() > 0:
            print('initializing...')
            time.sleep(2)
        while sim_all.getNumIncomplete() > 0 and sim_all.getRunSimulation():
            print('running...')
            time.sleep(2)

        status = sim_all.getNumError() + sim_all.getNumIncomplete()

        event_log = sym_prj.getEventLog()
        if 'Error' in event_log:
            status += 1

        sym_prj.regenerate()  # required to remove error jobs from jobmanager
        sym_prj.close()
        return status, event_log

    @unittest.skipIf(getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_run_job(self):
        """Run a distributable python function as Job
            This job is expected to succeed
        """
        sym_file = os.path.join(Paths.getSyMSpaceRoot(), 'System', 'Python', 'tests', 'test_symspace', 'test_jobinterface',
                                'projects', 'RunPythonJob', 'RunPythonJob.mop')

        status, event_log = self.run_sim(sym_file)

        if status == 0:
            print('Global Update completed successfully, as expected')
        else:
            print('Warning: Global Update incomplete')
            print(event_log)
            self.fail('Global Update incomplete')

    @unittest.skipIf(getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_run_job_variation(self):
        """Run a distributable python function with variation as multiple Jobs
            This jobs are expected to succeed
        """
        sym_file = os.path.join(Paths.getSyMSpaceRoot(), 'System', 'Python', 'tests', 'test_symspace', 'test_jobinterface',
                                'projects', 'RunPythonJobVariation', 'RunPythonJobVariation.mop')

        status, event_log = self.run_sim(sym_file)

        if status == 0:
            print('Global Update completed successfully, as expected')
        else:
            print('Warning: Global Update incomplete')
            print(event_log)
            self.fail('Global Update incomplete')

    @unittest.skipIf(getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_run_job_failing(self):
        """Run a distributable python function as Job
            This job is expected to fail, due to an exception raised in the run function
        """
        sym_file = os.path.join(Paths.getSyMSpaceRoot(), 'System', 'Python', 'tests', 'test_symspace', 'test_jobinterface',
                                'projects', 'RunPythonJobFail', 'RunPythonJobFail.mop')

        status, event_log = self.run_sim(sym_file)

        if status == 0:
            print('Warning: Global Update completed successfully')
            print(event_log)
            self.fail('The project was meant to fail, but did not!')
        else:
            print('Global Update incomplete, as expected')

    @unittest.skipIf(getInterpreter() == 'python', 'needs jep interpreter to run')
    def test_run_job_exitcode(self):
        """Run a distributable python function as Job
            This job is expected to fail, due to its non-zero exit code in the run function
        """
        sym_file = os.path.join(Paths.getSyMSpaceRoot(), 'System', 'Python', 'tests', 'test_symspace', 'test_jobinterface',
                                'projects', 'RunPythonJobExitCode', 'RunPythonJobExitCode.mop')

        status, event_log = self.run_sim(sym_file)

        if status == 0:
            print('Warning: Global Update completed successfully')
            print(event_log)
            self.fail('The project was meant to fail, but did not!')
        else:
            print('Global Update incomplete, as expected')


if __name__ == '__main__':
    from lcm.utils.testing import runUnittest

    runUnittest(getInterpreter())
