/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package jobmanager.server;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.Map;

import javax.swing.UIManager;

import jobmanager.JobDesc;
import jobmanager.JobType;
import jobmanager.server.gui.JobManagerView;

/**
 * Main class of job-manager
 * 
 * @author Werner Koppelsttter
 */
public class JobManager {

	/** View of JobManager */
	static JobManagerView view = null;

	/** SocketListener */
	private static SocketListener server;

	/** Statuslistener of jobs */
	private static JobStatusListener jobs;

	/**
	 * List of available clients (local, ...)
	 */
	private static Map<JobType, Client> clients = new Hashtable<JobType, Client>();

	/**
	 * Main function to start from command-line
	 * 
	 * @param args
	 *            Command-line parameters
	 * @throws MalformedURLException
	 */
	public static void main(String[] args) throws MalformedURLException {
		boolean loop;

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
		}

		try {
			JobManager.server = new SocketListener(JobDesc.getPort());
		} catch (IOException e) {
			System.exit(0);
		}

		// create job-listener
		JobManager.jobs = new JobStatusListener();

		JobManager.clients.put(JobType.LOCAL, jobmanager.server.local.LocalClient.getInstance());

		try {
			JobManager.clients.put(JobType.CLUSTER, (Client) Class
					.forName("jobmanager.server.htcondor.HTCondorInterface").getMethod("getClient").invoke(null));
		} catch (Throwable ex) {
		}

		// exit if no view and no job is in JobManager
		do {
			try {
				Thread.sleep(10000);
			} catch (InterruptedException e) {
			}
			loop = (JobManager.view != null) || (Job.joblist.size() != 0) || (JobListener.count != 0);
		} while (loop);

		System.exit(0);
	}

	/**
	 * Get used port
	 * 
	 * @return Used port
	 */
	public static int getPort() {
		return JobManager.server.getPort();
	}

	/**
	 * Show or Hide JobManger-gui
	 * 
	 * @param visible
	 *            State for visible
	 */
	public static synchronized void setVisible(boolean visible) {

		if (visible) {
			if (JobManager.view == null) {
				JobManager.view = new JobManagerView();
			}
		} else {
			if (JobManager.view != null) {
				JobManager.view.close();
				JobManager.view = null;
			}
		}
	}

	/**
	 * Get selected client
	 * 
	 * @param type
	 *            Type of job
	 * @return Client-object
	 */
	public static Client getClient(JobType type) {
		return JobManager.clients.get(type);
	}

	/**
	 * Refresh jobmanager
	 */
	public static void refresh() {
		JobManager.jobs.interrupt();
		for (JobType key : JobManager.clients.keySet()) {
			JobManager.clients.get(key).refresh();
		}
	}

	/**
	 * Reset to add new jobs (new iteration)
	 */
	public static void resetAdd() {
		for (JobType key : JobManager.clients.keySet()) {
			JobManager.clients.get(key).resetAdd();
		}
	}

	/**
	 * Refresh lists (Add remove external jobs)
	 */
	public static void refreshList() {
		for (JobType key : JobManager.clients.keySet()) {
			JobManager.clients.get(key).refreshList();
		}
	}

}
