/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package purge;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import javax.swing.WindowConstants;

/**
 * Purge ProEngineer directory. Remove suffix number if possible
 * 
 * @author Werner Koppelsttter
 */
public class Purge {

	/**
	 * Flag for cleaning logs
	 */
	protected boolean clean = false;

	/**
	 * Flag for virtual purge
	 */
	protected boolean virtual = false;

	/**
	 * Flag for log purge
	 */
	protected boolean log = false;

	/**
	 * List of warnings
	 */
	protected LinkedList<String> warnings = new LinkedList<String>();

	/**
	 * Purge directory. ussage: purge.jar [<directory<] [-clean] [-virtual] [-log]
	 * 
	 * @param args
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		File wd = null;
		Purge p;

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
		}

		// evaluate parameter
		if (args.length == 0) {
			System.out.println("USAGE:");
			System.out.println("\tpurge.jar [<directory>] [-clean] [-vitual]");
			System.out.println("\t\t-clean\tclean temporary files e.g. log, idx, m_p, lst");
			System.out.println("\t\t-virtual\tvirtual purge");
			System.out.println();
		} else {
			p = new Purge();
			for (String str : args) {
				if (str.compareToIgnoreCase("-clean") == 0) {
					p.setClean(true);
				} else if (str.compareToIgnoreCase("-virtual") == 0) {
					p.setVirtual(true);
				} else if (str.compareToIgnoreCase("-log") == 0) {
					p.setLog(true);
				} else {
					wd = new File(str);
				}
			}
			// get working dir
			if ((wd == null) || !wd.exists()) {
				wd = new File(System.getProperty("user.dir"));
			}
			p.purge(wd);
		}
	}

	/**
	 * Get virtual flag
	 * 
	 * @return Virtual flag
	 */
	public boolean getVirtual() {
		return this.virtual;
	}

	/**
	 * Set virtual flag
	 * 
	 * @param virtual
	 *            Virtual flag
	 */
	public void setVirtual(boolean virtual) {
		this.virtual = virtual;
	}

	/**
	 * Get log flag
	 * 
	 * @return Log flag
	 */
	public boolean getLog() {
		return this.log;
	}

	/**
	 * Set log flag
	 * 
	 * @param log
	 *            Log flag
	 */
	public void setLog(boolean log) {
		this.log = log;
	}

	/**
	 * Get clean-log flag
	 * 
	 * @return Clean-log flag
	 */
	public boolean getClean() {
		return this.clean;
	}

	/**
	 * Set clean-log flag
	 * 
	 * @param clean
	 *            Clean-log flag
	 */
	public void setClean(boolean clean) {
		this.clean = clean;
	}

	/**
	 * Purge directory
	 * 
	 * @param workdir
	 *            Working directory
	 */
	public void purge(File workdir) {
		JFrame frame;
		JList<String> list;
		JButton button;
		JPanel panel, buttonPanel;

		if (this.virtual) {
			this.warnings.add("VIRTUAL PURGE!");
		}

		// reduce Files
		this.reduceFiles(workdir);

		// clean logs
		if (this.clean) {
			this.cleanTmpFiles(workdir);
		}

		// display warnings
		frame = new JFrame("Purge warnings");
		frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		frame.setLayout(new BorderLayout());

		panel = new JPanel();
		frame.add(panel, BorderLayout.CENTER);
		panel.setLayout(new BorderLayout());

		list = new JList<String>(this.warnings.toArray(new String[0]));
		panel.add(new JScrollPane(list), BorderLayout.CENTER);

		buttonPanel = new JPanel();
		buttonPanel.setLayout(new BorderLayout());
		panel.add(buttonPanel, BorderLayout.SOUTH);

		button = new JButton("OK");
		button.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent arg0) {
				System.exit(0);
			}
		});
		buttonPanel.add(button, BorderLayout.CENTER);
		buttonPanel.add(new JPanel(), BorderLayout.NORTH);

		frame.add(new JPanel(), BorderLayout.WEST);
		frame.add(new JPanel(), BorderLayout.EAST);
		frame.add(new JPanel(), BorderLayout.NORTH);
		frame.add(new JPanel(), BorderLayout.SOUTH);

		frame.setLocation(200, 200);
		frame.setSize(500, 400);
		frame.setVisible(true);

	}

	/**
	 * Reduce files with a same name and diffrent numeric ending
	 * 
	 * @param workdir
	 *            Working directory
	 */
	protected void reduceFiles(File workdir) {
		File file, target = null, source = null;
		File[] foi = null;
		Hashtable<String, LinkedList<Integer>> list;
		Integer[] indizes;
		boolean fail = false;

		// find files of interest (with ".number")
		foi = workdir.listFiles(new FileFilter() {
			@Override
			public boolean accept(File file) {
				boolean accept = false;
				if (file.isFile()) {
					if (file.isFile() && file.getName().matches(".+\\.\\d+")) {
						accept = true;
					}
				}
				return accept;
			}
		});

		// group list and reduce groups
		list = this.groupFiles(foi);
		for (String path : list.keySet()) {
			indizes = list.get(path).toArray(new Integer[0]);
			Arrays.sort(indizes);
			fail = false;
			for (int k = 0; k < (indizes.length - 1); k++) {
				file = new File(path + "." + indizes[k]);
				if (file.canWrite()) {
					if (this.log) {
						this.warnings.add("Delete: '" + file.getName() + "'");
					}
					if (!this.virtual) {
						file.delete();
					}
				} else {
					fail = true;
					this.warnings.add("'" + file.getName() + "': Unable to delete file (write-protected)");
				}
			}

			// rename active file (delete source only if rename possible
			// (subversion))
			target = new File(path);
			source = new File(path + "." + indizes[indizes.length - 1]);
			if (fail) {
				this.warnings.add("'" + source.getName() + "': Unable to rename file (older file exist)");
			} else {
				if (target.exists() && !target.canWrite()) {
					this.warnings.add("'" + target.getName() + "': Unable to delete file (write-protected)");
				}
				if (!source.canWrite()) {
					this.warnings.add("'" + source.getName() + "': Unable to rename file (write-protected)");
				}

				// delete target
				if (target.exists() && target.canWrite() && source.canWrite()) {
					if (this.log) {
						this.warnings.add("Delete: '" + target.getName() + "'");
					}
					if (!this.virtual) {
						target.delete();
					}
				}

				// rename source
				if (!target.exists() && source.canWrite()) {
					if (this.log) {
						this.warnings.add("Rename: '" + source.getName() + "'");
					}
					if (!this.virtual) {
						source.renameTo(target);
					}
				}
			}
		}
	}

	/**
	 * Delete defined files (log, idx, m_p, lst)
	 * 
	 * @param workdir
	 *            working directory
	 */
	protected void cleanTmpFiles(File workdir) {
		File[] foi = null;

		// find logs
		foi = workdir.listFiles(new FileFilter() {
			@Override
			public boolean accept(File file) {
				boolean accept = false;
				if (file.isFile()) {
					accept |= file.getName().endsWith(".log");
					accept |= file.getName().endsWith(".idx");
					accept |= file.getName().endsWith(".m_p");
					accept |= file.getName().endsWith(".lst");
				}
				return accept;
			}
		});

		// delete logs
		for (File file : foi) {
			if (file.canWrite()) {
				if (this.log) {
					this.warnings.add("Delete: '" + file.getName() + "'");
				}
				if (!this.virtual) {
					file.delete();
				}
			} else {
				this.warnings.add("'" + file.getName() + "': Unable to delete log-file (write-protected)");
			}
		}
	}

	/**
	 * Group files with a same name and different numeric ending
	 * 
	 * @param files
	 *            files to group
	 * @return grouped list of files
	 */
	protected Hashtable<String, LinkedList<Integer>> groupFiles(File[] files) {
		Hashtable<String, LinkedList<Integer>> list = new Hashtable<String, LinkedList<Integer>>();
		LinkedList<Integer> group;
		String path;
		Integer fileIndex;
		int index;

		for (File f : files) {
			// split to path and file-index
			path = f.getPath();
			index = path.lastIndexOf('.');
			fileIndex = Integer.valueOf(path.substring(index + 1));
			path = path.substring(0, index);

			// add to list
			group = list.get(path);
			if (group == null) {
				group = new LinkedList<Integer>();
				list.put(path, group);
			}
			group.add(fileIndex);
		}

		return list;
	}

}
